#!/usr/bin/python3

import os
import textwrap
from unittest.mock import patch

from osbuild.testutil import make_fake_tree

STAGE_NAME = "org.osbuild.zipl.inst"


@patch("subprocess.run")
def test_zipl_inst_default(mocked_run, tmp_path, stage_module):
    make_fake_tree(tmp_path, {
        "/boot/loader/entries/ostree-1-fedora-coreos.conf": textwrap.dedent("""\
        title Fedora CoreOS 39.20240206.dev.0 (ostree:0)
        version 1
        options ignition.platform.id=qemu ostree=/ostree/boot.1/fedora-coreos/abcd/0
        linux /vmlinuz-6.6.14-200.fc39.s390x
        initrd /initramfs-6.6.14-200.fc39.s390x.img
        """),
        "/vmlinuz-6.6.14-200.fc39.s390x": "fake-vmlinuz",
        "/initramfs-6.6.14-200.fc39.s390x.img": "fake-initramfs.img",
    })
    fake_paths = {
        "mounts": os.fspath(tmp_path),
        "devices": "/run/osbuild/dev",
    }
    fake_devices = {
        "disk": {
            "path": "some-device-path",
        }
    }
    fake_options = {
        "kernel": "1",
        "location": "some-location",
    }

    stage_module.main(fake_paths, fake_devices, fake_options)
    assert len(mocked_run.call_args_list) == 1
    args, kwargs = mocked_run.call_args_list[0]
    assert len(args) == 1
    run_argv = args[0]
    assert "ignition.platform.id=qemu ostree=/ostree/boot.1/fedora-coreos/abcd/0" in run_argv
    assert kwargs["check"]


@patch("subprocess.run")
def test_zipl_inst_kernel_opts_append(mocked_run, tmp_path, stage_module):
    make_fake_tree(tmp_path, {
        "/boot/loader/entries/00_linux.conf": textwrap.dedent("""\
        title fake-linux
        linux /vmlinuz
        initrd /initrd.img
        options root=/dev/sda1 ro
        version 6.8
        """),
        "/vmlinuz": "fake-vmlinuz",
        "/initrd.img": "fake-initrd.img",
    })
    fake_paths = {
        "mounts": os.fspath(tmp_path),
        "devices": "/run/osbuild/dev",
    }
    fake_devices = {
        "disk": {
            "path": "some-device-path",
        }
    }
    fake_options = {
        "kernel": "6.8",
        "location": "some-location",
        "kernel_opts_append": ["some=more", "opts"],
    }

    stage_module.main(fake_paths, fake_devices, fake_options)
    assert len(mocked_run.call_args_list) == 1
    args, kwargs = mocked_run.call_args_list[0]
    assert len(args) == 1
    run_argv = args[0]
    assert "root=/dev/sda1 ro some=more opts" in run_argv
    assert kwargs["check"]



#!/usr/bin/python3

import os

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.idmap"


@pytest.mark.parametrize(
    "test_data",
    [
        # #####################
        # idmapping by id lists
        {
            "items": {}
        },
        {
            "items": {
                "/var/test-1": [
                    "b:9999:0:10"
                ],
            }
        },
        {
            "items": {
                "/var/test-1": [
                    "u:9999:0:10"
                ],
            }
        },
        {
            "items": {
                "/var/test-1": [
                    "g:9999:0:10"
                ],
            }
        },
        {
            "items": {
                "/var/test-1": [
                    "9999:0:10"
                ],
            }
        },
        {
            "items": {
                "/var/test-1": [
                    "555:444:10"
                ],
            }
        },
        {
            "items": {
                "/var/test-1": [
                    "555:444:10", "g:9999:0:10", "u:8888:0:10"
                ],
            }
        },
        # ########################
        # idmap by user/group name
        {
            "items": {
                "/var/test-1": {
                    "user": "test-user",
                    "group": "test-group"
                }
            }
        },
    ],
)
@pytest.mark.parametrize("stage_schema", ["1"], indirect=True)
def test_schema_validation_good(stage_schema, test_data):
    test_input = {
        "name": STAGE_NAME,
        "options": test_data,
    }
    res = stage_schema.validate(test_input)
    assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"


@pytest.mark.parametrize(
    "test_data,expected_err",
    [
        # #####################
        # idmapping by id lists

        # Missing slash as first character
        (
            {
                "items": {
                    "var/test-1": [],
                }
            },
            "does not match any of the regexes:",
        ),
        # Invalid idmap 1 - 9 instead of g
        (
            {
                "items": {
                    "/var/test-1": [
                        "9:9999:0:10"
                    ],
                }
            },
            "is not valid under any of the given schemas",
        ),
        # Invalid idmap 2 - empty idmap
        (
            {
                "items": {
                    "/var/test-1": [
                        ""
                    ],
                }
            },
            "is not valid under any of the given schemas",
        ),
        # Invalid idmap 3 - empty idmap
        (
            {
                "items": {
                    "/var/test-1": [
                        "blobb:1000:0:1"
                    ],
                }
            },
            "is not valid under any of the given schemas",
        ),
        # Invalid idmap 4 - negative values
        (
            {
                "items": {
                    "/var/test-1": [
                        "blobb:1000:0:-1"
                    ],
                }
            },
            "is not valid under any of the given schemas",
        ),

        # ########################
        # idmap by user/group name
        (
            {
                "items": {
                    "/var/test-1": {}
                }
            },
            "{} is not valid under any of the given schemas",
        ),
        (
            {
                "items": {
                    "/var/test-1": {
                        "user": "test-user",
                    }
                }
            },
            "{'user': 'test-user'} is not valid under any of the given schemas",
        ),
        (
            {
                "items": {
                    "/var/test-1": {
                        "group": "test-group",
                    }
                }
            },
            "{'group': 'test-group'} is not valid under any of the given schemas",
        ),
        (
            {
                "items": {
                    "/var/test-1": {
                        "user": "",
                        "group": "",
                    }
                }
            },
            "{'user': '', 'group': ''} is not valid under any of the given schemas",
        ),
    ],
)
@pytest.mark.parametrize("stage_schema", ["1"], indirect=True)
def test_schema_validation_bad(stage_schema, test_data, expected_err):
    test_input = {
        "name": STAGE_NAME,
        "options": test_data,
    }
    res = stage_schema.validate(test_input)
    assert res.valid is False
    testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


def compare_mappings(stage_module, mappings, expected_uid_maps, expected_gid_maps):
    user_mappings = mappings[stage_module.IDType.USER]
    assert len(user_mappings) == len(expected_uid_maps)
    for i, umap in enumerate(user_mappings):
        expected = expected_uid_maps[i]

        assert umap.inside == expected[0], f"Inside value: Expected {expected[0]}, got {umap.inside}"
        assert umap.outside == expected[1], f"Outside value: Expected {expected[1]}, got {umap.outside}"
        assert umap.size == expected[2], f"Size value: Expected {expected[2]}, got {umap.size}"

    group_mappings = mappings[stage_module.IDType.GROUP]
    assert len(group_mappings) == len(expected_gid_maps)
    for i, gmap in enumerate(group_mappings):
        expected = expected_gid_maps[i]

        assert gmap.inside == expected[0], f"Inside value: Expected {expected[0]}, got {gmap.inside}"
        assert gmap.outside == expected[1], f"Outside value: Expected {expected[1]}, got {gmap.outside}"
        assert gmap.size == expected[2], f"Size value: Expected {expected[2]}, got {gmap.size}"


@pytest.mark.parametrize(
    # The same order in idmap_inputs and expected_ is assumed by the test
    # which needs to be taken into account writing the test data
    "idmap_inputs,expected_uid_maps,expected_gid_maps", [
        ([], [], []),
        (["u:1000:0:10"], [(1000, 0, 10)], []),
        (["g:1000:0:10"], [], [(1000, 0, 10)]),
        (["b:1000:0:10"], [(1000, 0, 10)], [(1000, 0, 10)]),
        (["1000:0:10"], [(1000, 0, 10)], [(1000, 0, 10)]),
        (["b:1000:0:10", "u:1020:20:10"], [(1000, 0, 10), (1020, 20, 10)], [(1000, 0, 10)]),
    ]
)
def test_idmaps_from_list_good(stage_module, idmap_inputs, expected_uid_maps, expected_gid_maps):
    mappings = stage_module.IDMap.from_list(idmap_inputs)
    compare_mappings(stage_module, mappings, expected_uid_maps, expected_gid_maps)


@pytest.mark.parametrize(
    "idmap_inputs", [
        ([""]),
        (["x:100:0:10"]),
        (["u:100:0:10", "u1000:0:10"]),
        (["u:-100:0:10"]),
    ]
)
def test_idmaps_from_list_bad(stage_module, idmap_inputs):
    with pytest.raises(ValueError):
        stage_module.IDMap.from_list(idmap_inputs)


SAMPLE_ETC_SUBUID = """
test-user:1000:3000
blobb:10000:5000
"""
SAMPLE_ETC_SUBGID = """
test-group:1000:3000
blobb:10000:5000
"""


@pytest.mark.parametrize(
    "user,group,subuid_content,subgid_content,expected_uid_maps,expected_gid_maps", [
        ("", "", SAMPLE_ETC_SUBUID, SAMPLE_ETC_SUBGID, [], []),
        ("", "", SAMPLE_ETC_SUBUID, SAMPLE_ETC_SUBGID, [], []),
        ("test-user", "", SAMPLE_ETC_SUBUID, SAMPLE_ETC_SUBGID, [(1000, 0, 3000)], []),
        ("", "test-group", SAMPLE_ETC_SUBUID, SAMPLE_ETC_SUBGID, [], [(1000, 0, 3000)]),
        ("test-user", "test-group", SAMPLE_ETC_SUBUID, SAMPLE_ETC_SUBGID, [(1000, 0, 3000)], [(1000, 0, 3000)]),
        ("blobb", "blobb", SAMPLE_ETC_SUBUID, SAMPLE_ETC_SUBGID, [(10000, 0, 5000)], [(10000, 0, 5000)]),
    ]
)
def test_idmaps_from_subordinate_files_good(
        tmp_path,
        stage_module,
        user,
        group,
        subuid_content,
        subgid_content,
        expected_uid_maps,
        expected_gid_maps):

    # Set up test data
    os.makedirs(f"{tmp_path}/etc")
    with open(f"{tmp_path}/etc/subuid", "w", encoding="utf-8") as f:
        f.write(subuid_content)
    with open(f"{tmp_path}/etc/subgid", "w", encoding="utf-8") as f:
        f.write(subgid_content)

    mappings = stage_module.IDMap.from_subordinate_files(tmp_path, user, group)
    compare_mappings(stage_module, mappings, expected_uid_maps, expected_gid_maps)


@pytest.mark.parametrize(
    "fuid,fgid,idmap_inputs,expected", [
        (0, 0, ["u:1000:0:10"], (1000, 0)),
        (0, 0, ["g:1000:0:10"], (0, 1000)),
        (0, 0, ["b:1000:0:10"], (1000, 1000)),
        (0, 0, ["1000:0:10"], (1000, 1000)),
        (3, 3, ["b:1000:0:10"], (1003, 1003)),
        (11, 11, ["b:1000:0:10"], (65534, 65534)),
    ]
)
def test_determine_new_ids(stage_module, fuid, fgid, idmap_inputs, expected):
    idmaps = stage_module.IDMap.from_list(idmap_inputs)

    new_ids = stage_module.determine_new_ids(fuid, fgid, idmaps)
    assert new_ids == expected



#!/usr/bin/python3

from unittest.mock import call, patch

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.ostree.config"


@pytest.mark.parametrize("test_data,expected", [
    ({"sysroot": {"readonly": True}}, ["sysroot.readonly", "true"]),
    ({"sysroot": {"readonly": False}}, ["sysroot.readonly", "false"]),
    ({"sysroot": {"bootloader": "grub2"}}, ["sysroot.bootloader", "grub2"]),
    ({"sysroot": {"bootprefix": True}}, ["sysroot.bootprefix", "true"]),
    ({"sysroot": {"bootprefix": False}}, ["sysroot.bootprefix", "false"]),
    ({"sysroot": {"bls-append-except-default": "foobar"}}, ["sysroot.bls-append-except-default", "foobar"]),
],)
@patch("osbuild.util.ostree.cli")
def test_ostree_config(mock_ostree_cli, stage_module, test_data, expected):
    options = {
        "repo": "/ostree/repo",
        "config": {}
    }
    options["config"].update(test_data)
    stage_module.main("/run/osbuild/tree", options)
    repo = "/run/osbuild/tree/ostree/repo"
    assert mock_ostree_cli.call_args_list == [
        call("config", "set", *expected, repo=repo)]


@pytest.mark.parametrize("test_data,expected_err", [
    ({"sysroot": {"readonly": "albert"}},
     "'albert' is not of type 'boolean'"),
    ({"sysroot": {"bootloader": "lilo"}},
     "'lilo' is not one of ['none', 'auto', 'grub2', 'syslinux', 'uboot', 'zipl', 'aboot']"),
    ({"sysroot": {"bootprefix": "php"}},
     "'php' is not of type 'boolean'"),
])
def test_schema_validation_ostree_config(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {
            "repo": "/ostree/repo",
            "config": {}
        },
    }
    test_input["options"]["config"].update(test_data)
    res = stage_schema.validate(test_input)
    assert res.valid is False
    testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)



#!/usr/bin/python3

import subprocess
import sys
from unittest.mock import call, patch

import pytest

from osbuild import testutil

TEST_INPUT = [
    (
        [
            {"var": "ssh_idle_timeout_value", "value": 60},
        ], [
            "--var-value", "ssh_idle_timeout_value=60",
        ],
    ),
    (
        [
            {"var": "ssh_idle_timeout_value", "value": "60"},
        ], [
            "--var-value", "ssh_idle_timeout_value=60",
        ],
    ),
    (
        [
            {"var": "var1", "value": "val1"},
            {"var": "var2", "value": "val2"}
        ], [
            "--var-value", "var1=val1", "--var-value", "var2=val2",
        ],
    ),
]


STAGE_NAME = "org.osbuild.oscap.autotailor"


@pytest.fixture(name="fake_input")
def fake_input_fixture():
    return {
        "name": STAGE_NAME,
        "options": {
            "filepath": "/some/filepath",
            "config": {
                "profile_id": "some-profile-id",
                "datastream": "some-datastream",
                "tailored_profile_id": "some-new-profile",
            }
        },
    }


@pytest.fixture(name="fake_json_input")
def fake_json_input_fixture():
    return {
        "name": STAGE_NAME,
        "options": {
            "filepath": "tailoring-output.xml",
            "config": {
                "datastream": "some-datastream",
                "tailored_profile_id": "some-new-profile",
                "tailoring_file": "tailoring-file.json",
            }
        },
    }


@pytest.mark.parametrize("test_overrides,expected", TEST_INPUT)
@patch("subprocess.run")
def test_oscap_autotailor_overrides_smoke(mock_subprocess_run, fake_input, stage_module, test_overrides, expected):
    options = fake_input["options"]
    options["config"]["overrides"] = test_overrides
    stage_module.main("/some/sysroot", options)

    assert mock_subprocess_run.call_args_list == [
        call(["/usr/bin/autotailor", "--output", "/some/sysroot/some/filepath",
              "--new-profile-id", "some-new-profile"] +
             expected +
             ["some-datastream", "some-profile-id"],
             encoding='utf8', stdout=sys.stderr, check=True)]


@pytest.mark.parametrize(
    "test_data,expected_err",
    [
        ({
            "overrides": [
                {
                    "no": "var",
                    "value": 50,
                },
            ]
        }, "{'profile_id': 'some-profile-id', 'datastream': 'some-datastream', 'tailored_profile_id': 'some-new-profile',"
            + " 'overrides': [{'no': 'var', 'value': 50}]} is not valid under any of the given schemas"),
        ({
            "overrides": [
                {
                    "no": "value",
                    "var": "some",
                },
            ]
        }, "{'profile_id': 'some-profile-id', 'datastream': 'some-datastream', 'tailored_profile_id': 'some-new-profile',"
            + " 'overrides': [{'no': 'value', 'var': 'some'}]} is not valid under any of the given schemas"),
        ({
            "overrides": [
                {
                    "var": "ssh_idle_timeout_value",
                    "value": {"some": "object"},
                },
            ]
        }, "{'profile_id': 'some-profile-id', 'datastream': 'some-datastream', 'tailored_profile_id': 'some-new-profile',"
            + " 'overrides': [{'var': 'ssh_idle_timeout_value', 'value': {'some': 'object'}}]} is not valid under any of the given schemas"),
    ],
)
@pytest.mark.parametrize("stage_schema", ["1"], indirect=True)
def test_schema_validation_oscap_autotailor(fake_input, stage_schema, test_data, expected_err):
    fake_input["options"]["config"].update(test_data)
    res = stage_schema.validate(fake_input)

    assert res.valid is False
    testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


@patch("subprocess.run")
def test_oscap_autotailor_json_smoke(mock_subprocess_run, fake_json_input, stage_module):
    options = fake_json_input["options"]
    stage_module.main("/some/sysroot", options)

    assert mock_subprocess_run.call_args_list == [
        call(["/usr/bin/autotailor", "--output", "/some/sysroot/tailoring-output.xml",
              "--new-profile-id", "some-new-profile",
              "--json-tailoring", "/some/sysroot/tailoring-file.json", "some-datastream"],
             encoding='utf8', stdout=sys.stderr, check=True)]


@pytest.mark.parametrize(
    "test_data,expected_err",
    [
        ({}, "{} is not valid under any of the given schemas"),
        ({
            "tailored_profile_id": "some-new-profile"
        }, "{'tailored_profile_id': 'some-new-profile'}"
            + " is not valid under any of the given schemas"),
        ({
            "datastream": "some-datastream",
            "tailored_profile_id": "some-new-profile"
        }, "{'datastream': 'some-datastream', 'tailored_profile_id': 'some-new-profile'}"
            + " is not valid under any of the given schemas"),
        ({
            "datastream": "some-datastream",
            "profile_id": "some-profile-id",
        }, "{'datastream': 'some-datastream', 'profile_id': 'some-profile-id'}"
            + " is not valid under any of the given schemas"),
        ({
            "datastream": "some-datastream",
            "tailoring_file": "/some/tailoring/file.json"
        }, "{'datastream': 'some-datastream', 'tailoring_file': '/some/tailoring/file.json'}"
            + " is not valid under any of the given schemas"),
    ],
)
@pytest.mark.parametrize("stage_schema", ["1"], indirect=True)
def test_schema_validation_oscap_json_autotailor(fake_json_input, stage_schema, test_data, expected_err):
    fake_json_input["options"]["config"] = test_data
    res = stage_schema.validate(fake_json_input)
    assert res.valid is False
    testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


JSON_TAILORING = """{
  "profiles": [
    {
      "id": "some-profile-id",
      "base_profile_id": "some-profile-id",
      "rules": {
        "some-rule": {
          "evaluate": true,
          "severity": "high"
        }
      },
      "variables": {
        "some-variable": {
          "value": 600
        }
      }
    }
  ]
}
"""


@pytest.mark.parametrize(
    "expected_profile",
    [
        ("xccdf_org.ssgproject.content_profile_some-new-profile"),
        ("xccdf_org.ssgproject.content_profile_some-other-profile")
    ]
)
@pytest.mark.skipif(not testutil.has_executable("autotailor"), reason="no autotailor executable")
def test_oscap_autotailor_json_profile_override(fake_json_input, stage_module, expected_profile, tmp_path):
    options = fake_json_input["options"]
    options["config"]["tailored_profile_id"] = expected_profile

    results_file = tmp_path / options["filepath"]
    tailoring_file = tmp_path / options["config"]["tailoring_file"]
    tailoring_file.write_text(JSON_TAILORING)

    stage_module.main(str(tmp_path), options)

    result = subprocess.run(
        ["oscap", "info", "--profiles", results_file],
        stdout=subprocess.PIPE,
        check=True,
        encoding="utf-8",
    )

    assert f"Id: {expected_profile}\n" in result.stdout



#!/usr/bin/python3
import os

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.dnf4.versionlock"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({}, "'add' is a required property"),
    # good
    ({"add": ["shim-x64-*"]}, ""),
    ({"add": ["proto-1:1.1", "deftero-0:2.2", "trito-3:3.3-3.fc33"]}, ""),
])
def test_schema_validation_dnf4_versionlock(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": test_data
    }
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


@pytest.mark.parametrize("test_data", [
    ({"add": ["shim-x64-*"]}),
    ({"add": ["proto-1:1.1", "deftero-0:2.2", "trito-3:3.3-3.fc33"]}),
])
def test_locklist_dnf4_versionlock(tmp_path, stage_module, test_data):
    os.environ["SOURCE_DATE_EPOCH"] = "1554721380"
    plugins_dir = os.path.join(tmp_path, "etc/dnf/plugins/")
    locklist_path = os.path.join(plugins_dir, "versionlock.list")
    os.makedirs(plugins_dir)
    stage_module.main(tmp_path, test_data)

    with open(locklist_path, mode="r", encoding="utf-8") as locklist_fp:
        locklist_data = locklist_fp.readlines()

    for idx, package in enumerate(test_data["add"]):
        assert locklist_data[idx * 3] == "\n"
        assert locklist_data[idx * 3 + 1] == "# Added lock on Mon Apr  8 11:03:00 2019\n"
        assert locklist_data[idx * 3 + 2] == package + "\n"



"""Defines standard-agnostic data model for an SBOM."""

import abc
import urllib.parse
import uuid
from datetime import datetime
from enum import Enum, auto
from typing import Dict, List, Optional, Set


class ChecksumAlgorithm(Enum):
    SHA1 = auto()
    SHA224 = auto()
    SHA256 = auto()
    SHA384 = auto()
    SHA512 = auto()
    MD5 = auto()


class BasePackage(abc.ABC):
    """Represents a software package."""

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        name: str,
        version: str,
        filename: str = "",
        license_declared: str = "",
        vendor: str = "",
        checksums: Optional[Dict[ChecksumAlgorithm, str]] = None,
        homepage: str = "",
        download_url: str = "",
        build_date: Optional[datetime] = None,
        summary: str = "",
        description: str = "",
        depends_on: Optional[Set["BasePackage"]] = None,
        optional_depends_on: Optional[Set["BasePackage"]] = None,
    ) -> None:
        self.name = name
        self.version = version
        self.filename = filename
        self.license_declared = license_declared
        self.vendor = vendor
        self.checksums = checksums or {}
        self.homepage = homepage
        self.download_url = download_url
        self.build_date = build_date
        self.summary = summary
        self.description = description
        self.depends_on = depends_on or set()
        self.optional_depends_on = optional_depends_on or set()

    @abc.abstractmethod
    def uuid(self) -> str:
        """
        Returns a stable UUID for the package.
        """

    @abc.abstractmethod
    def source_info(self) -> str:
        """
        Return a string describing the source of the package.
        """

    @abc.abstractmethod
    def purl(self) -> str:
        """
        Return a Package URL for the package.

        The PURL format is:
        pkg:<type>/<namespace>/<name>@<version>?<qualifiers>#<subpath>

        Core PURL spec is defined at:
        https://github.com/package-url/purl-spec/blob/master/PURL-SPECIFICATION.rst
        """


class RPMDependency:
    """Represents an RPM dependency or provided capability."""

    def __init__(self, name: str, relation: str = "", version: str = "") -> None:
        self.name = name
        self.relation = relation
        self.version = version

    def __str__(self) -> str:
        return f"{self.name} {self.relation} {self.version}"


class RPMPackage(BasePackage):
    """Represents an RPM package."""

    def __init__(
        self,
        name: str,
        version: str,
        release: str,
        architecture: str,
        epoch: int = 0,
        filename: str = "",
        license_declared: str = "",
        vendor: str = "",
        checksums: Optional[Dict[ChecksumAlgorithm, str]] = None,
        homepage: str = "",
        download_url: str = "",
        build_date: Optional[datetime] = None,
        summary: str = "",
        description: str = "",
        depends_on: Optional[Set["BasePackage"]] = None,
        optional_depends_on: Optional[Set["BasePackage"]] = None,
        repository_url: str = "",
        source_rpm: str = "",
        rpm_provides: Optional[List[RPMDependency]] = None,
        rpm_requires: Optional[List[RPMDependency]] = None,
        rpm_recommends: Optional[List[RPMDependency]] = None,
        rpm_suggests: Optional[List[RPMDependency]] = None,
    ) -> None:
        super().__init__(
            name,
            version,
            filename,
            license_declared,
            vendor,
            checksums,
            homepage,
            download_url,
            build_date,
            summary,
            description,
            depends_on,
            optional_depends_on,
        )
        self.release = release
        self.architecture = architecture
        self.epoch = epoch
        self.repository_url = repository_url
        self.source_rpm = source_rpm
        self.rpm_provides = rpm_provides or []
        self.rpm_requires = rpm_requires or []
        self.rpm_recommends = rpm_recommends or []
        self.rpm_suggests = rpm_suggests or []

    def source_info(self) -> str:
        """
        Return a string describing the source of the RPM package.
        """
        if self.source_rpm:
            return f"Source RPM: {self.source_rpm}"
        return ""

    def uuid(self) -> str:
        """
        Returns a stable UUID for the same RPM package as defined by the PURL.
        """
        return str(uuid.uuid3(uuid.NAMESPACE_URL, self._purl(with_repo_url=False)))

    def _purl(self, with_repo_url=True) -> str:
        """
        Return a Package URL for the RPM package.

        Optionally don't include the repository URL in the PURL. This is useful
        to generate a PURL that can be used to identify the same package, regardless
        of the repository it was found in.

        PURL spec for RPMs is defined at:
        https://github.com/package-url/purl-spec/blob/master/PURL-TYPES.rst#rpm
        """
        namespace = ""
        if self.vendor:
            namespace = f"{urllib.parse.quote(self.vendor.lower())}/"

        purl = f"pkg:rpm/{namespace}{self.name}@{self.version}-{self.release}?arch={self.architecture}"

        if self.epoch:
            purl += f"&epoch={self.epoch}"

        if with_repo_url and self.repository_url:
            # https://github.com/package-url/purl-spec/blob/master/PURL-SPECIFICATION.rst#character-encoding
            purl += f"&repository_url={urllib.parse.quote(self.repository_url, safe='/:=')}"

        return purl

    def purl(self) -> str:
        return self._purl()



import os
import pathlib
from types import ModuleType

import pytest

from osbuild import mounts, testutil
from osbuild.testutil.imports import import_module_from_path


@pytest.fixture(name="mounts_module")
def mounts_module_fixture(request: pytest.FixtureRequest) -> ModuleType:
    """mounts_module is a fixture that imports a stage module by its name
    defined in MOUNTS_NAME in the test module.
    """
    if not hasattr(request.module, "MOUNTS_NAME"):
        raise ValueError("mounts_module fixture must be used in a test module that defines MOUNTS_NAME")

    mounts_name = request.module.MOUNTS_NAME
    caller_dir = pathlib.Path(request.node.fspath).parent
    module_path = caller_dir.parent / mounts_name
    return import_module_from_path("mounts", os.fspath(module_path))


@pytest.fixture
def mounts_service(mounts_module) -> ModuleType:
    """mounts_service is a fixture that imports a mounts module by its name
    defined in MOUNTS_NAME in the test module and returns a MountsService
    """
    service_cls = testutil.find_one_subclass_in_module(mounts_module, mounts.MountService)
    fd = testutil.make_fake_service_fd()
    srv_obj = service_cls.from_args(["--service-fd", str(fd)])
    return srv_obj



#!/usr/bin/python3

from unittest.mock import patch

import pytest

STAGE_NAME = "org.osbuild.dracut"


@pytest.mark.parametrize("with_initoverlayfs,expected_argv2", [
    (False, "/usr/bin/dracut"),
    (True, "/usr/bin/initoverlayfs-install"),
])
@patch("subprocess.run")
def test_dracut_with_initoverlayfs(mocked_run, tmp_path, stage_module, with_initoverlayfs, expected_argv2):
    options = {
        "kernel": [
            "5.14.0-247.el9.x86_64"
        ],
        "initoverlayfs": with_initoverlayfs,
    }

    stage_module.main(str(tmp_path), options)

    # We expect 7 calls to run(): 3 mount + chroot + 3 umount
    assert len(mocked_run.call_args_list) == 7
    args, kwargs = mocked_run.call_args_list[3]  # chroot is the 4th call
    assert kwargs.get("check") is True
    run_argv = args[0]
    assert run_argv[0] == "chroot"
    assert run_argv[2] == expected_argv2


@patch("subprocess.run")
def test_dracut_logger(mocked_run, tmp_path, stage_module):
    fake_logger_path = tmp_path / "usr/bin/logger"
    fake_logger_path.parent.mkdir(parents=True)
    fake_logger_path.write_text("")
    fake_true_path = tmp_path / "usr/bin/true"
    fake_true_path.write_text("")

    options = {
        "kernel": [
            "5.14.0-247.el9.x86_64",
        ]
    }
    stage_module.main(tmp_path.as_posix(), options)
    assert len(mocked_run.call_args_list) == 9
    args, kwargs = mocked_run.call_args_list[0]  # bind-mount is the 1th call
    assert kwargs.get("check") is True
    run_argv = args[0]
    assert run_argv == [
        "mount", "--rbind",
        fake_true_path.as_posix(), fake_logger_path.as_posix(),
    ]
    args, kwargs = mocked_run.call_args_list[8]  # umount is the 9th call
    assert kwargs.get("check") is False
    run_argv = args[0]
    assert run_argv == ["umount", fake_logger_path.as_posix()]



"""Wrapper module for output formatting."""

import sys
from typing import Dict


class VT:
    """Video terminal output, disables formatting when stdout is not a tty."""

    isatty: bool

    escape_sequences: Dict[str, str] = {
        "reset": "\033[0m",

        "bold": "\033[1m",

        "red": "\033[31m",
        "green": "\033[32m",
    }

    def __init__(self) -> None:
        self.isatty = sys.stdout.isatty()

    def __getattr__(self, name: str) -> str:
        if not self.isatty:
            return ""

        return self.escape_sequences[name]


fmt = VT()



#!/usr/bin/python3

import re

import pytest

from osbuild.testutil import (
    assert_jsonschema_error_contains,
)

STAGE_NAME = "org.osbuild.rhsm"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({
        "dnf-plugins": {
            "random": {
                "enabled": False,
            },
            "subscription-manager": {
                "enabled": False,
            },
        },
    }, r"Additional properties are not allowed \('random' was unexpected\)"),
    ({
        "dnf-plugins": {
            "product-id": {
                "enhanced": False,
            },
            "subscription-manager": {
                "enabled": False,
            },
        },
    }, r"Additional properties are not allowed \('enhanced' was unexpected\)"),
    ({
        "subscription-manager": {
            "rhsm": {
                "manage_repos": 45,
                "auto_enable_yum_plugins": False
            },
            "rhsmcertd": {
                "auto_registration": False,
            },
        }
    }, r"45 is not of type 'boolean'"),
    ({
        "subscription-manager": {
            "rhsm": {
                "random": False
            },
            "rhsmcertd": {
                "auto_registration": False,
            },
        }
    }, r"Additional properties are not allowed \('random' was unexpected\)"),
    ({
        "subscription-manager": {
            "random": {},
        }
    }, r"Additional properties are not allowed \('random' was unexpected\)"),
    ({
        "subscription-manager": {
            "rhsmcertd": {
                "random": False,
            },
        }
    }, r"Additional properties are not allowed \('random' was unexpected\)"),
    # good
    ({}, ""),
    ({
        "dnf-plugins": {},
        "yum-plugins": {},
        "subscription-manager": {},
    }, ""),
    ({
        "dnf-plugins": {
            "product-id": {
                "enabled": False,
            },
            "subscription-manager": {
                "enabled": False,
            },
        },
        "yum-plugins": {
            "product-id": {
                "enabled": False,
            },
            "subscription-manager": {
                "enabled": False,
            },
        },
        "subscription-manager": {
            "rhsm": {
                "manage_repos": False,
                "auto_enable_yum_plugins": False,
            },
            "rhsmcertd": {
                "auto_registration": False,
            },
        },
    }, ""),
    ({
        "dnf-plugins": {
            "product-id": {
                "enabled": True,
            },
            "subscription-manager": {
                "enabled": True,
            },
        },
        "yum-plugins": {
            "product-id": {
                "enabled": True,
            },
            "subscription-manager": {
                "enabled": True,
            },
        },
        "subscription-manager": {
            "rhsm": {
                "manage_repos": True,
                "auto_enable_yum_plugins": True,
            },
            "rhsmcertd": {
                "auto_registration": True,
            },
        },
    }, ""),
])
def test_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {},
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        assert_jsonschema_error_contains(res, re.compile(expected_err), expected_num_errs=1)



#!/usr/bin/python3

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.wsl-distribution.conf"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({"oobe": "must-be-object"}, "'must-be-object' is not of type 'object'"),
    ({"shortcut": "must-be-object"}, "'must-be-object' is not of type 'object'"),
    ({"shortcut": {"enabled": 1}}, "1 is not of type 'boolean'"),
    ({"oobe": {"default_uid": True}}, "True is not of type 'integer'"),
    # good
    ({
        "oobe":
        {
            "default_uid": 1000,
            "default_name": "RedHatEnterpriseLinux-10.0",
        },
        "shortcut":
        {
            "enabled": True,
            "icon": "/usr/share/pixmaps/fedora-logo.ico",
        }
    }, "")
])
def test_wsl_distribution_conf_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


def test_wsl_distribution_conf_writes_file(tmp_path, stage_module):
    treepath = tmp_path / "tree"

    etcpath = treepath / "etc"
    etcpath.mkdir(parents=True, exist_ok=True)

    confpath = "wsl-distribution.conf"

    fullpath = etcpath / confpath

    options = {
        "oobe": {
            "default_name": "RedHatEnterpriseLinux-10.0",
            "default_uid": 1000,
        },
        "shortcut": {
            "enabled": True,
            "icon": "/usr/share/pixmaps/fedora-logo.ico",
        }
    }

    stage_module.main(treepath, options)

    assert fullpath.exists()

    confdata = fullpath.read_text()

    assert confdata == """\
[oobe]
defaultUid = 1000
defaultName = RedHatEnterpriseLinux-10.0

[shortcut]
enabled = true
icon = /usr/share/pixmaps/fedora-logo.ico"""



import os
import subprocess


class Chroot:
    """
    Sets up mounts for the virtual filesystems inside a root tree, preparing it for running commands using chroot. This
    should be used whenever a stage needs to run a command against the root tree but doesn't support a --root option or
    similar.
    Cleans up mounts when done.

    This mounts /proc, /dev, and /sys.
    """

    def __init__(self, root: str, bind_mounts=None):
        self.root = root
        self._bind_mounts = bind_mounts or []

    def __enter__(self):
        for d in ["/proc", "/dev", "/sys"]:
            if not os.path.exists(self.root + d):
                print(f"Making missing chroot directory: {d}")
                os.makedirs(self.root + d)

        subprocess.run(["mount", "-t", "proc", "-o", "nosuid,noexec,nodev",
                        "proc", f"{self.root}/proc"],
                       check=True)

        try:
            subprocess.run(["mount", "-t", "devtmpfs", "-o", "mode=0755,noexec,nosuid,strictatime",
                            "devtmpfs", f"{self.root}/dev"],
                           check=True)
        except subprocess.CalledProcessError:
            # If we're not allowed to mount a new devtmpfs, try bind-mounting the host /dev
            subprocess.run(["mount", "--bind", "/dev", f"{self.root}/dev"],
                           check=True)

        subprocess.run(["mount", "-t", "sysfs", "-o", "nosuid,noexec,nodev",
                        "sysfs", f"{self.root}/sys"],
                       check=True)

        for d in self._bind_mounts:
            target_path = os.path.join(self.root, d.lstrip("/"))
            if not os.path.exists(target_path):
                print(f"Making missing chroot directory: {d}")
                os.makedirs(target_path)
            subprocess.run(["mount", "--rbind", d, target_path], check=True)

        return self

    def __exit__(self, exc_type, exc_value, tracebk):
        failed_umounts = []
        for d in ["/proc", "/dev", "/sys"]:
            if subprocess.run(["umount", "--lazy", self.root + d], check=False).returncode != 0:
                failed_umounts.append(d)
        for d in self._bind_mounts[::-1]:
            target_path = os.path.join(self.root, d.lstrip("/"))
            if subprocess.run(["umount", "--lazy", target_path], check=False).returncode != 0:
                failed_umounts.append(d)
        if failed_umounts:
            print(f"Error unmounting paths from chroot: {failed_umounts}")

    def run(self, cmd, **kwargs):
        cmd = ["chroot", self.root] + cmd
        # pylint: disable=subprocess-run-check
        return subprocess.run(cmd, **kwargs)  # noqa: PLW1510



import os.path
import pathlib
import platform
import shutil
import subprocess
import sys
from contextlib import contextmanager


def ldconfig(*dirs):
    # ld.so.conf must exist, or `ldconfig` throws a warning
    subprocess.run(["touch", "/etc/ld.so.conf"], check=True)

    if len(dirs) > 0:
        with open("/etc/ld.so.conf", "w", encoding="utf8") as f:
            for d in dirs:
                f.write(f"{d}\n")
            f.flush()

    subprocess.run(["ldconfig"], check=True)


def sysusers():
    try:
        subprocess.run(
            ["systemd-sysusers"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=True,
        )
    except subprocess.CalledProcessError as error:
        sys.stderr.write(error.stdout)
        sys.exit(1)


@contextmanager
def create_machine_id_if_needed(tree="", keep_empty=False):
    """Create a machine-id with a fake machine id if it does not exist.
       The machine-id file will be delete at context exit unless specified
       with 'keep_empty' variable. In that case an empty machine-id will
       be kept.
    """
    path = pathlib.Path(f"{tree}/etc/machine-id")
    try:
        if not path.exists():
            path.parent.mkdir(mode=0o755, exist_ok=True)
            with path.open(mode="w", encoding="utf8") as f:
                # create a fake machine ID to improve reproducibility
                f.write("ffffffffffffffffffffffffffffffff\n")
                path.chmod(0o444)
        yield
    finally:
        path.unlink()
        if keep_empty:
            path.touch()
            path.chmod(0o444)


def tmpfiles():
    # Allow systemd-tmpfiles to return non-0. Some packages want to create
    # directories owned by users that are not set up with systemd-sysusers.
    subprocess.run(["systemd-tmpfiles", "--create"], check=False)


def nsswitch():
    # the default behavior is fine, but using nss-resolve does not
    # necessarily work in a non-booted container, so make sure that
    # is not configured.
    try:
        os.remove("/etc/nsswitch.conf")
    except FileNotFoundError:
        pass


def python_alternatives():
    """/usr/bin/python3 is a symlink to /etc/alternatives/python3, which points
    to /usr/bin/python3.6 by default. Recreate the link in /etc, so that
    shebang lines in stages and assemblers work.
    """
    os.makedirs("/etc/alternatives", exist_ok=True)
    try:
        os.symlink("/usr/bin/python3.6", "/etc/alternatives/python3")
    except FileExistsError:
        pass


def sequoia():
    # This provides a default set of crypto-policies which is important for
    # re-enabling SHA1 support with rpm (so we can cross-build CentOS-Stream-9
    # images).
    os.makedirs("/etc/crypto-policies", exist_ok=True)
    shutil.copytree(
        "/usr/share/crypto-policies/back-ends/DEFAULT", "/etc/crypto-policies/back-ends"
    )


def quirks():
    # Platform specific quirks
    env = os.environ.copy()

    if platform.machine() == "aarch64":
        # Work around a bug in qemu-img on aarch64 that can lead to qemu-img
        # hangs when more then one coroutine is use (which is the default)
        # See https://bugs.launchpad.net/qemu/+bug/1805256
        env["OSBUILD_QEMU_IMG_COROUTINES"] = "1"

    return env


def minimal_passwd():
    # ensure minimal /etc/passwd, "podman mount" needs that
    passwd_path = pathlib.Path("/etc/passwd")
    if not passwd_path.exists():
        passwd_path.write_text("root:x:0:0:root:/root:/bin/bash")



#!/usr/bin/python3

import os
import pathlib
from unittest.mock import patch

import pytest

import osbuild.meta
from osbuild import testutil

MOUNTS_NAME = "org.osbuild.ddi"


def _fake_args(source="/dev/loop0", target="/mnt/image", root="/tmp/buildroot", options=None):
    args = {
        "source": source,
        "target": target,
        "root": root,
        "options": options or {},
    }
    return args


@patch("subprocess.run")
def test_ddi_mount_simple(mocked_run, mounts_service):
    args = _fake_args()
    result = mounts_service.mount(args)

    assert result == os.path.join(args["root"], args["target"].lstrip("/"))
    assert len(mocked_run.call_args_list) == 1
    run_args, run_kwargs = mocked_run.call_args_list[0]
    assert run_args[0] == [
        "systemd-dissect", "--mount", "--mkdir",
        "/dev/loop0", os.path.join("/tmp/buildroot", "mnt/image"),
    ]
    assert run_kwargs["check"] is True


@patch("subprocess.run")
def test_ddi_mount_all_options(mocked_run, mounts_service):
    args = _fake_args(options={
        "growfs": False,
        "read-only": True,
        "fsck": False,
        "discard": "all",
        "image-policy": "root=verity",
        "image-filter": "usr",
    })
    mounts_service.mount(args)

    assert len(mocked_run.call_args_list) == 1
    run_args, _ = mocked_run.call_args_list[0]
    cmd = run_args[0]
    assert cmd == [
        "systemd-dissect", "--mount", "--mkdir",
        "--growfs", "no",
        "--read-only",
        "--fsck", "no",
        "--discard", "all",
        "--image-policy", "root=verity",
        "--image-filter", "usr",
        "/dev/loop0", os.path.join("/tmp/buildroot", "mnt/image"),
    ]


@patch("subprocess.run")
def test_ddi_mount_readonly_only(mocked_run, mounts_service):
    args = _fake_args(options={"read-only": True})
    mounts_service.mount(args)

    run_args, _ = mocked_run.call_args_list[0]
    cmd = run_args[0]
    assert cmd == [
        "systemd-dissect", "--mount", "--mkdir",
        "--read-only",
        "/dev/loop0", os.path.join("/tmp/buildroot", "mnt/image"),
    ]


@patch("subprocess.run")
def test_ddi_mount_defaults_not_emitted(mocked_run, mounts_service):
    args = _fake_args(options={"growfs": True, "fsck": True, "read-only": False})
    mounts_service.mount(args)

    run_args, _ = mocked_run.call_args_list[0]
    cmd = run_args[0]
    assert cmd == [
        "systemd-dissect", "--mount", "--mkdir",
        "/dev/loop0", os.path.join("/tmp/buildroot", "mnt/image"),
    ]


@patch("subprocess.run")
def test_ddi_umount(mocked_run, mounts_service):
    mounts_service.mountpoint = "/tmp/buildroot/mnt/image"
    mounts_service.umount()

    assert len(mocked_run.call_args_list) == 1
    run_args, run_kwargs = mocked_run.call_args_list[0]
    assert run_args[0] == ["systemd-dissect", "--umount", "/tmp/buildroot/mnt/image"]
    assert run_kwargs.get("check") is False
    assert mounts_service.mountpoint is None


@patch("subprocess.run")
def test_ddi_umount_noop_when_not_mounted(mocked_run, mounts_service):
    mounts_service.umount()
    assert len(mocked_run.call_args_list) == 0


@patch("subprocess.run")
def test_ddi_umount_not_called_twice(mocked_run, mounts_service):
    mounts_service.mountpoint = "/tmp/buildroot/mnt/image"
    mounts_service.umount()
    mounts_service.umount()
    assert len(mocked_run.call_args_list) == 1


@pytest.mark.parametrize("test_data,expected_err", [
    # bad - missing required fields
    ({}, "'name' is a required property"),
    ({}, "'target' is a required property"),
    ({}, "'source' is a required property"),
    # bad - invalid discard value
    ({"name": "ddi", "source": "/img", "target": "/mnt", "options": {"discard": "invalid"}},
     "'invalid' is not one of"),
    # bad - extra option
    ({"name": "ddi", "source": "/img", "target": "/mnt", "options": {"bogus": True}},
     "Additional properties are not allowed"),
    # good - minimal
    ({"name": "ddi", "source": "/img", "target": "/mnt"}, ""),
    # good - all options
    ({"name": "ddi", "source": "/img", "target": "/mnt", "options": {
        "growfs": False, "read-only": True, "fsck": False,
        "discard": "loop", "image-policy": "root=verity", "image-filter": "usr",
    }}, ""),
])
def test_schema_validation(test_data, expected_err):
    root = pathlib.Path(__file__).parent.parent.parent
    mod_info = osbuild.meta.ModuleInfo.load(root, "Mount", MOUNTS_NAME)
    schema = osbuild.meta.Schema(mod_info.get_schema(), MOUNTS_NAME)
    test_input = {
        "type": MOUNTS_NAME,
        "options": {}
    }
    test_input.update(test_data)
    res = schema.validate(test_input)
    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err)



# pylint: disable=too-many-branches
# pylint: disable=too-many-nested-blocks

import itertools
import os
import os.path
import tempfile
from typing import Any, Dict, List, Optional, Set

import dnf
import hawkey
import libdnf
from dnf.i18n import ucd

import osbuild.solver.model as model
from osbuild.solver import SolverBase, modify_rootdir_path, read_keys
from osbuild.solver.exceptions import DepsolveError, MarkingError, NoReposError, RepoError
from osbuild.solver.request import DepsolveCmdArgs, SearchCmdArgs, SolverConfig
from osbuild.util.sbom.dnf import dnf_pkgset_to_sbom_pkgset
from osbuild.util.sbom.spdx import sbom_pkgset_to_spdx2_doc


def _reldep_to_dependency(reldep: hawkey.Reldep) -> model.Dependency:
    """
    Convert a hawkey.Reldep to an RPM Dependency.

    Note: Handles compatibility with RHEL-8 where Reldep objects don't have name/relation/version attributes.
    """
    try:
        return model.Dependency(reldep.name, reldep.relation.strip(), reldep.version)
    except AttributeError:
        # '_hawkey.Reldep' object has no attribute 'name' in the version shipped on RHEL-8
        dep_parts = str(reldep).split()
        while len(dep_parts) < 3:
            dep_parts.append("")
        return model.Dependency(dep_parts[0], dep_parts[1].strip(), dep_parts[2])


def _dnf_pkg_to_package(pkg: dnf.package.Package) -> model.Package:
    """
    Convert a dnf.package.Package to an RPM Package.
    """
    kwargs = {
        "name": pkg.name,
        "version": pkg.version,
        "release": pkg.release,
        "arch": pkg.arch,
        "epoch": int(pkg.epoch),
        "group": pkg.group or "",
        "download_size": pkg.downloadsize or 0,
        "install_size": pkg.installsize or 0,
        "license": pkg.license or "",
        "source_rpm": pkg.sourcerpm or "",
        "build_time": pkg.buildtime or 0,
        "packager": pkg.packager or "",
        "vendor": pkg.vendor or "",
        "url": pkg.url or "",
        "summary": pkg.summary or "",
        "description": pkg.description or "",
        "provides": sorted([_reldep_to_dependency(p) for p in pkg.provides], key=str),
        "requires": sorted([_reldep_to_dependency(p) for p in pkg.requires], key=str),
        "requires_pre": sorted([_reldep_to_dependency(p) for p in pkg.requires_pre], key=str),
        "conflicts": sorted([_reldep_to_dependency(p) for p in pkg.conflicts], key=str),
        "obsoletes": sorted([_reldep_to_dependency(p) for p in pkg.obsoletes], key=str),
        "regular_requires": sorted([_reldep_to_dependency(p) for p in pkg.regular_requires], key=str),
        "recommends": sorted([_reldep_to_dependency(p) for p in pkg.recommends], key=str),
        "suggests": sorted([_reldep_to_dependency(p) for p in pkg.suggests], key=str),
        "enhances": sorted([_reldep_to_dependency(p) for p in pkg.enhances], key=str),
        "supplements": sorted([_reldep_to_dependency(p) for p in pkg.supplements], key=str),
        "files": pkg.files,
        "location": pkg.location or "",
        "repo_id": pkg.repoid or "",
        "reason": pkg.reason or "",
    }
    # NB: prevent setting a [None] list for remote_locations
    remote_location = pkg.remote_location()
    if remote_location:
        kwargs["remote_locations"] = [remote_location]
    checksum = pkg.chksum
    if checksum and hawkey.chksum_name(checksum[0]) != "UNKNOWN":
        kwargs["checksum"] = model.Checksum(
            algorithm=hawkey.chksum_name(checksum[0]), value=checksum[1].hex())
    header_checksum = pkg.hdr_chksum
    if header_checksum and hawkey.chksum_name(header_checksum[0]) != "UNKNOWN":
        kwargs["header_checksum"] = model.Checksum(
            algorithm=hawkey.chksum_name(pkg.hdr_chksum[0]), value=pkg.hdr_chksum[1].hex())
    return model.Package(**kwargs)


def _dnf_repo_to_repository(
    repo: dnf.repo.Repo,
    root_dir: Optional[str],
    request_repo_ids: Set[str],
) -> model.Repository:
    """
    Convert a dnf.repo.Repo to a Repository.
    """
    return model.Repository(
        repo_id=repo.id,
        name=repo.name,
        baseurl=list(repo.baseurl),
        metalink=repo.metalink,
        mirrorlist=repo.mirrorlist,
        gpgcheck=repo.gpgcheck,
        repo_gpgcheck=repo.repo_gpgcheck,
        gpgkey=read_keys(repo.gpgkey, root_dir if repo.id not in request_repo_ids else None),
        sslverify=bool(repo.sslverify),
        sslcacert=repo.sslcacert,
        sslclientkey=repo.sslclientkey,
        sslclientcert=repo.sslclientcert,
        metadata_expire=f"{repo.metadata_expire}s" if repo.metadata_expire else None,
        module_hotfixes=repo.module_hotfixes,
        enabled=repo.enabled,
        priority=repo.priority,
        cost=repo.cost,
        skip_if_unavailable=repo.skip_if_unavailable,
        # username and password are intentionally omitted
    )


class DNF(SolverBase):
    SOLVER_NAME = "dnf"

    def __init__(
        self,
        config: SolverConfig,
        persistdir: os.PathLike,
        license_index_path: Optional[os.PathLike] = None,
    ):
        super().__init__(config, persistdir, license_index_path)

        self.repos = self.config.repos or []
        self.request_repo_ids = {repo.repo_id for repo in self.repos} if self.repos else set()
        self.root_dir = self.config.root_dir

        self.base = dnf.Base()

        # Enable fastestmirror to ensure we choose the fastest mirrors for
        # downloading metadata (when depsolving) and downloading packages.
        self.base.conf.fastestmirror = True

        # We use the same cachedir for multiple architectures. Unfortunately,
        # this is something that doesn't work well in certain situations
        # with zchunk:
        # Imagine that we already have cache for arch1. Then, we use dnf-json
        # to depsolve for arch2. If ZChunk is enabled and available (that's
        # the case for Fedora), dnf will try to download only differences
        # between arch1 and arch2 metadata. But, as these are completely
        # different, dnf must basically redownload everything.
        # For downloding deltas, zchunk uses HTTP range requests. Unfortunately,
        # if the mirror doesn't support multi range requests, then zchunk will
        # download one small segment per a request. Because we need to update
        # the whole metadata (10s of MB), this can be extremely slow in some cases.
        # I think that we can come up with a better fix but let's just disable
        # zchunk for now. As we are already downloading a lot of data when
        # building images, I don't care if we download even more.
        self.base.conf.zchunk = False

        # Set the rest of the dnf configuration.
        if self.config.module_platform_id:
            self.base.conf.module_platform_id = self.config.module_platform_id
        self.base.conf.config_file_path = "/dev/null"
        self.base.conf.persistdir = persistdir
        self.base.conf.cachedir = self.config.cachedir
        self.base.conf.substitutions['arch'] = self.config.arch
        self.base.conf.substitutions['basearch'] = dnf.rpm.basearch(self.config.arch)
        if self.config.releasever:
            self.base.conf.substitutions['releasever'] = self.config.releasever

        # variables substitution is only available when root_dir is provided
        if self.root_dir:
            # This sets the varsdir to ("{root_dir}/etc/yum/vars/", "{root_dir}/etc/dnf/vars/") for custom variable
            # substitution (e.g. CentOS Stream 9's $stream variable)
            self.base.conf.substitutions.update_from_etc(self.root_dir)

        if hasattr(self.base.conf, "optional_metadata_types") and self.config.optional_metadata:
            # the attribute doesn't exist on older versions of dnf; ignore the option when not available
            self.base.conf.optional_metadata_types.extend(self.config.optional_metadata)
        if self.config.proxy:
            self.base.conf.proxy = self.config.proxy

        try:
            for repo_conf in self.repos:
                self.base.repos.add(self._dnfrepo(repo_conf, self.base.conf, self.root_dir is not None))

            if self.root_dir:
                repos_dir = os.path.join(self.root_dir, "etc/yum.repos.d")
                self.base.conf.reposdir = repos_dir
                self.base.read_all_repos()
                for repo_id, repo_config in self.base.repos.items():
                    if repo_id not in self.request_repo_ids:
                        repo_config.sslcacert = modify_rootdir_path(repo_config.sslcacert, self.root_dir)
                        repo_config.sslclientcert = modify_rootdir_path(repo_config.sslclientcert, self.root_dir)
                        repo_config.sslclientkey = modify_rootdir_path(repo_config.sslclientkey, self.root_dir)

            self.base.update_cache()
            self.base.fill_sack(load_system_repo=False)
        except Exception as e:
            raise RepoError(e) from e

        if not self.base.repos._any_enabled():
            raise NoReposError("There are no enabled repositories")

        # enable module resolving
        self.base_module = dnf.module.module_base.ModuleBase(self.base)

    @staticmethod
    def _dnfrepo(desc: model.Repository, parent_conf=None, subs_links=False):
        """Makes a dnf.repo.Repo out of model.Repository configuration"""

        repo = dnf.repo.Repo(desc.repo_id, parent_conf)
        config = libdnf.conf.ConfigParser

        if desc.name:
            repo.name = desc.name

        def subs(basestr):
            if subs_links and parent_conf:
                return config.substitute(ucd(basestr), parent_conf.substitutions)
            return basestr

        # NB: the fact that at least one is set is checked when parsing the Solver request
        if desc.baseurl:
            repo.baseurl = [subs(url) for url in desc.baseurl]
        if desc.metalink:
            repo.metalink = subs(desc.metalink)
        if desc.mirrorlist:
            repo.mirrorlist = subs(desc.mirrorlist)

        if desc.sslverify is not None:
            repo.sslverify = desc.sslverify
        if desc.sslcacert:
            repo.sslcacert = desc.sslcacert
        if desc.sslclientkey:
            repo.sslclientkey = desc.sslclientkey
        if desc.sslclientcert:
            repo.sslclientcert = desc.sslclientcert

        if desc.gpgcheck is not None:
            repo.gpgcheck = desc.gpgcheck
        if desc.repo_gpgcheck is not None:
            repo.repo_gpgcheck = desc.repo_gpgcheck
        if desc.gpgkey:
            # gpgkey can contain a full key, or it can be a URL
            # dnf expects urls, so write the key to a temporary location and add the file://
            # path to repo.gpgkey
            keydir = os.path.join(parent_conf.persistdir, "gpgkeys")
            if not os.path.exists(keydir):
                os.makedirs(keydir, mode=0o700, exist_ok=True)

            for key in desc.gpgkey:
                if key.startswith("-----BEGIN PGP PUBLIC KEY BLOCK-----"):
                    # Not using with because it needs to be a valid file for the duration. It
                    # is inside the temporary persistdir so will be cleaned up on exit.
                    # pylint: disable=consider-using-with
                    keyfile = tempfile.NamedTemporaryFile(dir=keydir, delete=False)
                    keyfile.write(key.encode("utf-8"))
                    repo.gpgkey.append(f"file://{keyfile.name}")
                    keyfile.close()
                else:
                    repo.gpgkey.append(key)

        if desc.metadata_expire:
            repo.metadata_expire = desc.metadata_expire
        if desc.module_hotfixes is not None:
            repo.module_hotfixes = desc.module_hotfixes
        if desc.enabled is not None:
            repo.enabled = desc.enabled
        if desc.priority is not None:
            repo.priority = desc.priority
        if desc.cost is not None:
            repo.cost = desc.cost
        if desc.username:
            repo.username = desc.username
        if desc.password:
            repo.password = desc.password
        if desc.skip_if_unavailable is not None:
            repo.skip_if_unavailable = desc.skip_if_unavailable

        return repo

    def _sbom_for_pkgset(self, pkgset: List[dnf.package.Package]) -> Dict:
        """
        Create an SBOM document for the given package set.

        For now, only SPDX v2 is supported.
        """
        pkgset = dnf_pkgset_to_sbom_pkgset(pkgset)
        spdx_doc = sbom_pkgset_to_spdx2_doc(pkgset, self.license_index_path)
        return spdx_doc.to_dict()

    def _dnf_repo_to_repository(self, repo: dnf.repo.Repo) -> model.Repository:
        """
        Convert a dnf.repo.Repo to a Repository.
        """
        repo_model = _dnf_repo_to_repository(repo, self.root_dir, self.request_repo_ids)
        self.set_rhsm_flag(repo_model)
        return repo_model

    def dump(self) -> model.DumpResult:
        seen_repos = {}
        repositories = []

        def package_iter():
            for pkg in self.base.sack.query().available():
                if pkg.repo.id not in seen_repos:
                    repo = self._dnf_repo_to_repository(pkg.repo)
                    seen_repos[pkg.repo.id] = repo
                    repositories.append(repo)
                yield _dnf_pkg_to_package(pkg)

        return model.DumpResult(package_iter(), repositories)

    def search(self, args: SearchCmdArgs) -> model.SearchResult:
        """ Perform a search on the available packages"""
        packages = []
        repositories = {}

        # NOTE: Build query one piece at a time, don't pass all to filterm at the same
        # time.
        available = self.base.sack.query().available()
        for name in args.packages:
            # If the package name glob has * in it, use glob.
            # If it has *name* use substr
            # If it has neither use exact match
            if "*" in name:
                if name[0] != "*" or name[-1] != "*":
                    q = available.filter(name__glob=name)
                else:
                    q = available.filter(name__substr=name.replace("*", ""))
            else:
                q = available.filter(name__eq=name)

            if args.latest:
                q = q.latest()

            for pkg in q:
                packages.append(_dnf_pkg_to_package(pkg))
                if pkg.repo.id not in repositories:
                    repositories[pkg.repo.id] = self._dnf_repo_to_repository(pkg.repo)

        return model.SearchResult(packages, list(repositories.values()))

    def depsolve(self, args: DepsolveCmdArgs) -> model.DepsolveResult:
        """Perform a dependency resolution for the given transactions"""
        last_dnf_transaction: List[dnf.package.Package] = []
        # List of transaction results, each containing a list of packages that are a result of dependency resolution.
        # Each transaction result is a superset of the previous transaction result.
        # The package list in each transaction is alphabetically sorted by full NEVRA.
        transactions_results: List[List[model.Package]] = []
        repositories_by_id: Dict[str, model.Repository] = {}
        repositories: List[model.Repository] = []

        for transaction in args.transactions:
            self.base.reset(goal=True)
            self.base.sack.reset_excludes()

            # Restrict dependency resolution to only use packages from the
            # repos listed in repo_ids by excluding all packages from other
            # repos from the sack. The reponame parameter of install_specs()
            # only filters the explicitly requested packages, not their
            # dependencies.
            if transaction.repo_ids:
                allowed_repo_ids = set(transaction.repo_ids)
                for repo in self.base.repos.iter_enabled():
                    if repo.id not in allowed_repo_ids:
                        q = self.base.sack.query().available().filterm(reponame=repo.id)
                        self.base.sack.add_excludes(q)

            self.base.conf.install_weak_deps = transaction.install_weak_deps

            try:
                # set the packages from the last transaction as installed
                for installed_pkg in last_dnf_transaction:
                    self.base.package_install(installed_pkg, strict=True)

                # enabling a module means that packages can be installed from that
                # module
                if transaction.module_enable_specs:
                    self.base_module.enable(transaction.module_enable_specs)

                # installing a module takes the specification of the module and then
                # installs all packages belonging to its default group, modules to
                # install are listed directly in `package-specs` but prefixed with an
                # `@` *and* containing a `:` this is up to the user of the depsolver
                self.base.install_specs(
                    transaction.package_specs,
                    transaction.exclude_specs,
                )
            except dnf.exceptions.Error as e:
                raise MarkingError(e) from e

            try:
                self.base.resolve()
            except dnf.exceptions.Error as e:
                raise DepsolveError(e) from e

            transaction_result = []
            # store the current transaction result
            last_dnf_transaction.clear()
            for tsi in self.base.transaction:
                # NB: we don't use the 'install_set' helper, because it is a Python set() and it
                # does not guarantee a stable order.
                if tsi.action not in dnf.transaction.FORWARD_ACTIONS:
                    continue
                pkg = tsi.pkg
                last_dnf_transaction.append(pkg)
                transaction_result.append(_dnf_pkg_to_package(pkg))
                repo = pkg.repo
                if repo.id not in repositories_by_id:
                    repositories_by_id[repo.id] = self._dnf_repo_to_repository(repo)

            # NB: DNF4 solver returns packages in alphabetical order. However, to not depend on the DNF4 API
            # implementation details, we explicitly sort the packages alphabetically by full NEVRA.
            # NB: the org.osbuild.rpm stage as generated by osbuild/images does not depend on the order of packages,
            # because rpm gets the full package set at once and it will reorder the packages as needed when installing.
            transaction_result.sort()
            transactions_results.append(transaction_result)

        # NB: we sort the repositories by repo_id to ensure consistent ordering across DNF4 and DNF5.
        repositories = list(repositories_by_id.values())
        repositories.sort(key=lambda x: x.repo_id)

        sbom = None
        if args.sbom_request:
            sbom = self._sbom_for_pkgset(last_dnf_transaction)

        # if any modules have been requested we add sources for these so they can
        # be used by stages to enable the modules in the eventual artifact
        modules: Dict[str, Any] = {}

        for transaction in args.transactions:
            # module specifications must start with an "@", if they do we try to
            # ask DNF for a module by that name, if it doesn't exist it isn't a
            # module; otherwise it is and we should use it
            modules_in_package_specs = []

            for p in transaction.package_specs:
                if p.startswith("@") and self.base_module.get_modules(p):
                    modules_in_package_specs.append(p.lstrip("@"))

            if transaction.module_enable_specs or modules_in_package_specs:
                # we'll be checking later if any packages-from-modules are in the
                # packages-to-install set so let's do this only once here
                package_nevras = []

                # NB: len(transactions_results) == len(args.transactions), so transactions_results can't be empty.
                for package in transactions_results[-1]:
                    if package.epoch == 0:
                        package_nevras.append(
                            f"{package.name}-{package.version}-{package.release}.{package.arch}")
                    else:
                        package_nevras.append(
                            f"{package.name}-{package.epoch}:{package.version}-{package.release}.{package.arch}")

                for module_spec in itertools.chain(
                    transaction.module_enable_specs or [],
                    modules_in_package_specs,
                ):
                    module_packages, module_nsvcap = self.base_module.get_modules(module_spec)

                    # we now need to do an annoying dance as multiple modules could be
                    # returned by `.get_modules`, we need to select the *same* one as
                    # previously selected. we do this by checking if any of the module
                    # packages are in the packages set marked for installation.

                    # this is a result of not being able to get the enabled modules
                    # from the transaction, if that turns out to be possible then
                    # we can get rid of these shenanigans
                    for module_package in module_packages:
                        module_nevras = module_package.getArtifacts()

                        if any(module_nevra in package_nevras for module_nevra in module_nevras):
                            # a package from this module is being installed so we must
                            # use this module
                            module_ns = f"{module_nsvcap.name}:{module_nsvcap.stream}"

                            if module_ns not in modules:
                                modules[module_ns] = (module_package, set())

                            if module_nsvcap.profile:
                                modules[module_ns][1].add(module_nsvcap.profile)

                            # we are unable to skip the rest of the `module_packages`
                            # here since different profiles might be contained

        # now we have the information we need about modules so we need to return *some*
        # information to who is using the depsolver so they can use that information to
        # enable these modules in the artifact

        # there are two files that matter for each module that is used, the caller needs
        # to write a file to `/etc/dnf/modules.d/{module_name}.module` to enable the
        # module for dnf

        # the caller also needs to set up `/var/lib/dnf/modulefailsafe/` with the contents
        # of the modulemd for the selected modules, this is to ensure that even when a
        # repository is disabled or disappears that non-modular content can't be installed
        # see: https://dnf.readthedocs.io/en/latest/modularity.html#fail-safe-mechanisms
        modules_response = {}
        for module_ns, (module, profiles) in modules.items():
            modules_response[module.getName()] = {
                "module-file": {
                    "path": f"/etc/dnf/modules.d/{module.getName()}.conf",
                    "data": {
                        "name": module.getName(),
                        "stream": module.getStream(),
                        "profiles": list(profiles),
                        "state": "enabled",
                    }
                },
                "failsafe-file": {
                    "data": module.getYaml(),
                    "path": f"/var/lib/dnf/modulefailsafe/{module.getName()}:{module.getStream()}",
                },
            }

        return model.DepsolveResult(
            transactions=transactions_results,
            repositories=repositories,
            modules=modules_response if modules_response else None,
            sbom=sbom if sbom else None,
        )



#!/usr/bin/python3
"""
Lorax related utilities: Template parsing and execution

This module contains a re-implementation of the Lorax
template engine, but for osbuild. Not all commands in
the original scripting language are support, but all
needed to run the post install and cleanup scripts.
"""

import contextlib
import fnmatch
import os
import re
import shlex
import shutil
import subprocess
from typing import Any, Dict

import mako.template

from .replace import replace, rglob


class Script:

    # all built-in commands in a name to method map
    commands: Dict[str, Any] = {}

    # helper decorator to register builtin methods
    class command:
        def __init__(self, fn):
            self.fn = fn

        def __set_name__(self, owner, name):
            bultins = getattr(owner, "commands")
            bultins[name] = self.fn
            setattr(owner, name, self.fn)

    # Script class starts here
    def __init__(self, script, build, tree):
        self.script = script
        self.tree = tree
        self.build = build
        self._pkgnames = None

    def __call__(self):
        for i, line in enumerate(self.script):
            cmd, args = line[0], line[1:]
            ignore_error = False
            if cmd.startswith("-"):
                cmd = cmd[1:]
                ignore_error = True

            method = self.commands.get(cmd)

            if not method:
                raise ValueError(f"Unknown command: '{cmd}'")

            try:
                method(self, *args)
            except Exception:
                if ignore_error:
                    continue
                print(f"Error on line: {i} " + str(line))
                raise

    def tree_path(self, target):
        dest = os.path.join(self.tree, target.lstrip("/"))
        return dest

    def _all_pkgnames(self):
        """ Get the list of all package names installed
        On first call it runs rpm -qa and caches the results
        """
        if not self._pkgnames:
            cmd = ["rpm", "-qa", "--root", self.tree, "--qf=%{name}\n"]
            res = subprocess.run(cmd, stdout=subprocess.PIPE,
                                 check=True, encoding="utf8")
            self._pkgnames = sorted(res.stdout.splitlines())
        return self._pkgnames

    def _get_pkgfiles(self, pkg):
        """ Get a list of the files installed by a package """
        cmd = ["rpm", "-ql", "--root", self.tree, pkg]
        res = subprocess.run(cmd, stdout=subprocess.PIPE,
                             check=True, encoding="utf8")
        return sorted(res.stdout.splitlines())

    def _filelist(self, *pkg_specs):
        """ Return the list of files in the packages matching the globs """
        pkglist = []
        for pkg in self._all_pkgnames():
            if any(fnmatch.fnmatch(pkg, spec) for spec in pkg_specs):
                pkglist.append(pkg)

        # Return the files, not directories
        return set(f for pkg in pkglist for f in self._get_pkgfiles(pkg) if not os.path.isdir(self.tree_path(f)))

    @command
    def append(self, filename, data):
        target = self.tree_path(filename)
        dirname = os.path.dirname(target)
        os.makedirs(dirname, exist_ok=True)
        print(f"append '{target}' '{data}'")
        with open(target, "a", encoding="utf8") as f:
            f.write(bytes(data, "utf8").decode("unicode_escape"))
            f.write("\n")

    @command
    def mkdir(self, *dirs):
        for d in dirs:
            print(f"mkdir '{d}'")
            os.makedirs(self.tree_path(d), exist_ok=True)

    @command
    def move(self, src, dst):
        src = self.tree_path(src)
        dst = self.tree_path(dst)

        if os.path.isdir(dst):
            dst = os.path.join(dst, os.path.basename(src))

        print(f"move '{src}' -> '{dst}'")
        os.rename(src, dst)

    @command
    def install(self, src, dst):
        dst = self.tree_path(dst)
        for s in rglob(os.path.join(self.build, src.lstrip("/")), fatal=True):
            with contextlib.suppress(shutil.Error):
                print(f"install {s} -> {dst}")
                shutil.copy2(os.path.join(self.build, s), dst)

    @command
    def remove(self, *files):
        skiplist = ["/usr/lib/sysimage/rpm/", "/var/lib/rpm/", "/var/lib/yum", "/var/lib/dnf"]
        for g in files:
            # We need the rpm database in order to remove files from packages, so skip removing
            # it. This can be handled by skipping these paths in the squashfs/erofs stage.
            if any(g.startswith(p) for p in skiplist):
                print(f"remove: Skipping RPM database path {g}")
                continue

            for f in rglob(self.tree_path(g)):
                if os.path.isdir(f) and not os.path.islink(f):
                    shutil.rmtree(f)
                else:
                    os.unlink(f)
                print(f"remove '{f}'")

    @command
    def replace(self, pat, repl, *files):
        found = False
        for g in files:
            for f in rglob(self.tree_path(g)):
                found = True
                print(f"replace {f}: {pat} -> {repl}")
                replace(f, [(pat, repl)])

        if not found:
            assert found, f"No match for {pat} in {' '.join(files)}"

    @command
    def runcmd(self, *args):
        # We need /boot content in order to create the iso, so skip the commands acting on
        # boot. This can be handled by skipping these paths in the squashfs/erofs stage.
        if any("/boot" in a for a in args):
            print("runcmd: skipping operations on /boot")
            return

        print("run ", " ".join(args))
        subprocess.run(args, cwd=self.tree, check=True)

    @command
    def symlink(self, source, dest):
        target = self.tree_path(dest)
        if os.path.exists(target):
            self.remove(dest)
        print(f"symlink '{source}' -> '{target}'")
        os.symlink(source, target)

    @command
    def systemctl(self, verb, *units):
        assert verb in ('enable', 'disable', 'mask')
        self.mkdir("/run/systemd/system")
        cmd = ['systemctl', '--root', self.tree, '--no-reload', verb]

        for unit in units:
            with contextlib.suppress(subprocess.CalledProcessError):
                args = cmd + [unit]
                self.runcmd(*args)

    @command
    def removepkg(self, *pkgs):
        """
        removepkg PKGGLOB [PKGGLOB...]
          Delete the named package(s).

        IMPLEMENTATION NOTES:
          RPM scriptlets (%preun/%postun) are *not* run.
          Files are deleted, but directories are left behind.
        """
        for p in pkgs:
            filepaths = [f.lstrip('/') for f in self._filelist(p)]
            if filepaths:
                self.remove(*filepaths)
            else:
                print(f"removepkg {p}: no files to remove!")

    @command
    def removefrom(self, pkg, *globs):
        """
        removefrom PKGGLOB [--allbut] FILEGLOB [FILEGLOB...]
          Remove all files matching the given file globs from the package
          (or packages) named.
          If '--allbut' is used, all the files from the given package(s) will
          be removed *except* the ones which match the file globs.

          Examples:
            removefrom usbutils /usr/bin/*
            removefrom xfsprogs --allbut /bin/*
        """
        cmd = f"{pkg} {' '.join(globs)}"  # save for later logging
        keepmatches = False
        if globs[0] == '--allbut':
            keepmatches = True
            globs = globs[1:]
        # get pkg filelist and find files that match the globs
        filelist = self._filelist(pkg)
        matches = set()
        for g in globs:
            globs_re = re.compile(fnmatch.translate(g))
            m = [f for f in filelist if globs_re.match(f)]
            if m:
                matches.update(m)
            else:
                print(f"removefrom {pkg} {g}: no files matched!")
        if keepmatches:
            remove_files = filelist.difference(matches)
        else:
            remove_files = matches
        # remove the files
        if remove_files:
            self.remove(*remove_files)
        else:
            print(f"removefrom {cmd}: no files to remove!")

    # pylint: disable=anomalous-backslash-in-string
    @command
    def removekmod(self, *globs):
        '''
        removekmod GLOB [GLOB...] [--allbut] KEEPGLOB [KEEPGLOB...]
          Remove all files and directories matching the given file globs from the kernel
          modules directory.

          If '--allbut' is used, all the files from the modules will be removed *except*
          the ones which match the file globs. There must be at least one initial GLOB
          to search and one KEEPGLOB to keep. The KEEPGLOB is expanded to be *KEEPGLOB*
          so that it will match anywhere in the path.

          This only removes files from under /lib/modules/\\*/kernel/

          Examples:
            removekmod sound drivers/media drivers/hwmon drivers/video
            removekmod drivers/char --allbut virtio_console hw_random
        '''
        cmd = " ".join(globs)
        if "--allbut" in globs:
            idx = globs.index("--allbut")
            if idx == 0:
                raise ValueError("removekmod needs at least one GLOB before --allbut")

            # Apply keepglobs anywhere they appear in the path
            keepglobs = globs[idx + 1:]
            if len(keepglobs) == 0:
                raise ValueError("removekmod needs at least one GLOB after --allbut")

            globs = globs[:idx]
        else:
            # Nothing to keep
            keepglobs = []

        filelist = set()
        for g in globs:
            for top_dir in rglob(self.tree_path(f"/lib/modules/*/kernel/{g}")):
                for root, _dirs, files in os.walk(top_dir):
                    filelist.update(f"{root}/{f}" for f in files)

        # Remove anything matching keepglobs from the list
        matches = set()
        for g in keepglobs:
            globs_re = re.compile(fnmatch.translate(f"*{g}*"))
            m = [f for f in filelist if globs_re.match(f)]
            if m:
                matches.update(m)
            else:
                print(f"removekmod {g}: no files matched!")
        remove_files = filelist.difference(matches)

        if remove_files:
            list(os.unlink(f) for f in remove_files)
        else:
            print(f"removekmod {cmd}: no files to remove!")


def brace_expand(s):
    if not ('{' in s and ',' in s and '}' in s):
        return [s]

    result = []
    right = s.find('}')
    left = s[:right].rfind('{')
    prefix, choices, suffix = s[:left], s[left + 1:right], s[right + 1:]
    for choice in choices.split(','):
        result.extend(brace_expand(prefix + choice + suffix))

    return result


def brace_expand_line(line):
    return [after for before in line for after in brace_expand(before)]


def render_template(path, args):
    """Render a template at `path` with arguments `args`"""

    with open(path, "r", encoding="utf8") as f:
        data = f.read()

    tlp = mako.template.Template(text=data, filename=path)
    txt = tlp.render(**args)

    lines = map(lambda l: l.strip(), txt.splitlines())
    lines = filter(lambda l: l and not l.startswith("#"), lines)
    commands = map(shlex.split, lines)
    commands = map(brace_expand_line, commands)

    result = list(commands)
    return result



#!/usr/bin/python3

import textwrap

import pytest

STAGE_NAME = "org.osbuild.coreos.platform"


def make_fake_platforms_json(json_file_path):
    json_file_path.write_text(textwrap.dedent("""\
    {
      "only-grub": {
        "grub_commands": [
          "serial --speed=115200"
        ]
      },
      "azure": {
        "kernel_arguments": [
          "console=tty0",
          "console=ttyS0,115200n8"
        ]
      },
      "vmware": {
        "grub_commands": [
          "serial --speed=115200",
          "terminal_input serial console",
          "terminal_output serial console"
        ],
        "kernel_arguments": [
          "console=ttyS0,115200n8",
          "console=tty0"
        ]
      }
    }"""), encoding="utf8")
    return json_file_path


@pytest.mark.parametrize("platform,expected_grub,expected_kernel", [
    # no key is no error
    ("missing", [], []),
    # one of each
    ("only-grub", ["serial --speed=115200"], []),
    # examples from:
    # "podman run --rm quay.io/fedora/fedora-coreos:stable cat /usr/share/coreos-assembler/platforms.json"
    # thanks dusty!
    ("azure", [], ['console=tty0', 'console=ttyS0,115200n8']),
    ("vmware", ["serial --speed=115200", "terminal_input serial console",
                "terminal_output serial console"], ["console=ttyS0,115200n8", "console=tty0"]),
])
def test_process_platforms_json(tmp_path, stage_module, platform, expected_grub, expected_kernel):
    fake_platforms_path = make_fake_platforms_json(tmp_path / "platforms.json")
    grub_cmds, kernel_args = stage_module.process_platforms_json(fake_platforms_path, platform)
    assert grub_cmds == expected_grub
    assert kernel_args == expected_kernel


def test_generate_console_settings_none(tmp_path, stage_module):
    fake_console_settings_path = tmp_path / "boot/grub2/console.cfg"
    fake_console_settings_path.parent.mkdir(parents=True)
    console_settings = None

    stage_module.generate_console_settings_file(console_settings, fake_console_settings_path)
    # ensure we generate a template even if no settings are provided
    assert fake_console_settings_path.read_text(encoding="utf8") == textwrap.dedent("""\
    # Any non-default console settings will be inserted here.
    # CONSOLE-SETTINGS-START

    # CONSOLE-SETTINGS-END
    """)


def test_generate_console_settings_multiple(tmp_path, stage_module):
    fake_console_settings_path = tmp_path / "boot/grub2/console.cfg"
    fake_console_settings_path.parent.mkdir(parents=True)
    # console settings are the grub commands from the platforms.json
    console_settings = ["serial --speed=115200", "terminal_input serial console", "terminal_output serial console"]

    stage_module.generate_console_settings_file(console_settings, fake_console_settings_path)
    # ensure we generate a template even if no settings are provided
    assert fake_console_settings_path.read_text(encoding="utf8") == textwrap.dedent("""\
    # Any non-default console settings will be inserted here.
    # CONSOLE-SETTINGS-START
    serial --speed=115200
    terminal_input serial console
    terminal_output serial console
    # CONSOLE-SETTINGS-END
    """)



import collections
import contextlib
import hashlib
import json
import os
from fnmatch import fnmatch
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, Iterator, List, Optional

from . import buildroot, host, objectstore, remoteloop
from .api import API
from .devices import Device, DeviceManager
from .inputs import Input, InputManager
from .mounts import Mount, MountManager
from .objectstore import ObjectStore
from .qemu import Qemu
from .sources import Source, SourceResults
from .util import experimentalflags, osrelease

DEFAULT_CAPABILITIES = {
    "CAP_AUDIT_WRITE",
    "CAP_CHOWN",
    "CAP_DAC_OVERRIDE",
    "CAP_DAC_READ_SEARCH",
    "CAP_FOWNER",
    "CAP_FSETID",
    "CAP_IPC_LOCK",
    "CAP_LINUX_IMMUTABLE",
    "CAP_MAC_OVERRIDE",
    "CAP_MKNOD",
    "CAP_NET_BIND_SERVICE",
    "CAP_SETFCAP",
    "CAP_SETGID",
    "CAP_SETPCAP",
    "CAP_SETUID",
    "CAP_SYS_ADMIN",
    "CAP_SYS_CHROOT",
    "CAP_SYS_NICE",
    "CAP_SYS_RESOURCE"
}


def cleanup(*objs):
    """Call cleanup method for all objects, filters None values out"""
    _ = map(lambda o: o.cleanup(), filter(None, objs))


class StageResult:
    def __init__(self, origin: 'Stage', returncode: int, output: str, error: Dict[str, str]) -> None:
        self.name = origin.name
        self.id = origin.id
        self.success = returncode == 0
        self.output = output
        self.error = error

    def as_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "id": self.id,
            "success": self.success,
            "output": self.output,
            "error": self.error,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "StageResult":
        result = cls.__new__(cls)
        result.name = data["name"]
        result.id = data["id"]
        result.success = data["success"]
        result.output = data["output"]
        result.error = data["error"]
        return result


# pylint: disable=too-many-instance-attributes
class Stage:
    def __init__(self, info, source_options, build, base, options, source_epoch):
        self.info_name = info.name
        self.info_path = info.path
        self.info_caps = info.caps
        self.sources = source_options
        self.build = build
        self.base = base
        self.options = options
        self.source_epoch = source_epoch
        self.checkpoint = False
        self.inputs = {}
        self.devices = {}
        self.mounts = {}

    @property
    def name(self) -> str:
        return self.info_name

    @property
    def id(self) -> str:
        m = hashlib.sha256()
        m.update(json.dumps(self.name, sort_keys=True).encode())
        m.update(json.dumps(self.build, sort_keys=True).encode())
        m.update(json.dumps(self.base, sort_keys=True).encode())
        m.update(json.dumps(self.options, sort_keys=True).encode())
        if self.source_epoch is not None:
            m.update(json.dumps(self.source_epoch, sort_keys=True).encode())
        if self.inputs:
            data_inp = {n: i.id for n, i in self.inputs.items()}
            m.update(json.dumps(data_inp, sort_keys=True).encode())
        if self.mounts:
            data_mnt = [m.id for m in self.mounts.values()]
            m.update(json.dumps(data_mnt).encode())
        return m.hexdigest()

    @property
    def dependencies(self) -> Generator[str, None, None]:
        """Return a list of pipeline ids this stage depends on"""

        for ip in self.inputs.values():

            if ip.origin != "org.osbuild.pipeline":
                continue

            yield from ip.refs

    def add_input(self, name, info, origin, options=None):
        ip = Input(name, info, origin, options or {})
        self.inputs[name] = ip
        return ip

    def add_device(self, name, info, parent, options):
        dev = Device(name, info, parent, options)
        self.devices[name] = dev
        return dev

    def add_mount(self, name, info, device, partition, target, options):
        mount = Mount(name, info, device, partition, target, options)
        self.mounts[name] = mount
        return mount

    def to_dict(self, libdir):
        return {
            "info_name": self.info_name,
            "info_path": str(Path(self.info_path).relative_to(libdir)),
            "info_caps": list(self.info_caps),
            "sources": self.sources,
            "build": self.build,
            "base": self.base,
            "options": self.options,
            "source_epoch": self.source_epoch,
            "checkpoint": self.checkpoint,
            "inputs": {name: inp.to_dict(libdir) for name, inp in self.inputs.items()},
            "devices": {name: dev.to_dict(libdir) for name, dev in self.devices.items()},
            "mounts": {name: mnt.to_dict(libdir) for name, mnt in self.mounts.items()},
        }

    @classmethod
    def from_dict(cls, data, libdir):
        class InfoStub:
            def __init__(self, name: str, path: str, caps: set):
                self.name = name
                self.path = path
                self.caps = caps

        info = InfoStub(data["info_name"], Path(libdir) / data["info_path"], set(data["info_caps"]))
        stage = cls(
            info=info,
            source_options=data["sources"],
            build=data["build"],
            base=data["base"],
            options=data["options"],
            source_epoch=data["source_epoch"]
        )
        stage.checkpoint = data["checkpoint"]

        for name, inp_data in data["inputs"].items():
            inp = Input.from_dict(inp_data, libdir)
            stage.inputs[name] = inp

        devices_by_name = {}
        for name, dev_data in data["devices"].items():
            parent = devices_by_name.get(dev_data["parent_name"]) if dev_data["parent_name"] else None
            dev = Device.from_dict(dev_data, libdir, parent)
            stage.devices[name] = dev
            devices_by_name[name] = dev

        for name, mnt_data in data["mounts"].items():
            device = stage.devices.get(mnt_data["device_name"]) if mnt_data["device_name"] else None
            mnt = Mount.from_dict(mnt_data, libdir, device)
            stage.mounts[name] = mnt

        return stage

    def prepare_arguments(self, args, location):
        args["options"] = self.options
        args["meta"] = meta = {
            "id": self.id,
        }

        if self.source_epoch is not None:
            meta["source-epoch"] = self.source_epoch

        # Root relative paths: since paths are different on the
        # host and in the container they need to be mapped to
        # their path within the container. For all items that
        # have registered roots, re-root their path entries here
        for name, root in args.get("paths", {}).items():
            group = args.get(name)
            if not group or not isinstance(group, dict):
                continue
            for item in group.values():
                path = item.get("path")
                if not path:
                    continue
                item["path"] = os.path.join(root, path)

        with open(location, "w", encoding="utf-8") as fp:
            json.dump(args, fp)

    def run(
            self,
            tree_path,
            meta_name,
            runner,
            build_tree,
            store,
            monitor,
            libdir,
            debug_break="",
            timeout=None) -> StageResult:
        with contextlib.ExitStack() as cm:

            build_root = buildroot.BuildRoot(build_tree, runner.path, libdir, store.tmp)
            cm.enter_context(build_root)

            # if we have a build root, then also bind-mount the boot
            # directory from it, since it may contain efi binaries
            build_root.mount_boot = bool(self.build)

            # drop capabilities other than `DEFAULT_CAPABILITIES`
            build_root.caps = DEFAULT_CAPABILITIES | self.info_caps

            tmpdir = store.tempdir(prefix="buildroot-tmp-")
            tmpdir = cm.enter_context(tmpdir)

            inputs_tmpdir = os.path.join(tmpdir, "inputs")
            os.makedirs(inputs_tmpdir)
            inputs_mapped = "/run/osbuild/inputs"
            inputs: Dict[Any, Any] = {}

            devices_mapped = "/dev"
            devices: Dict[Any, Any] = {}

            mounts_tmpdir = os.path.join(tmpdir, "mounts")
            os.makedirs(mounts_tmpdir)
            mounts_mapped = "/run/osbuild/mounts"
            mounts: Dict[Any, Any] = {}

            os.makedirs(os.path.join(tmpdir, "api"))
            args_path = os.path.join(tmpdir, "api", "arguments")

            args = {
                "tree": "/run/osbuild/tree",
                "paths": {
                    "devices": devices_mapped,
                    "inputs": inputs_mapped,
                    "mounts": mounts_mapped,
                },
                "devices": devices,
                "inputs": inputs,
                "mounts": mounts,
            }

            ro_binds = [
                f"{self.info_path}:/run/osbuild/bin/{self.name}",
                f"{inputs_tmpdir}:{inputs_mapped}",
                f"{args_path}:/run/osbuild/api/arguments"
            ]

            binds = [
                tree_path + ":/run/osbuild/tree",
                meta_name + ":/run/osbuild/meta",
                f"{mounts_tmpdir}:{mounts_mapped}"
            ]

            storeapi = objectstore.StoreServer(store)
            cm.enter_context(storeapi)

            mgr = host.ServiceManager(monitor=monitor)
            cm.enter_context(mgr)

            ipmgr = InputManager(mgr, storeapi, inputs_tmpdir)
            for key, ip in self.inputs.items():
                data_inp = ipmgr.map(ip)
                inputs[key] = data_inp

            devmgr = DeviceManager(mgr, build_root.dev, tree_path)
            for name, dev in self.devices.items():
                devices[name] = devmgr.open(dev)

            mntmgr = MountManager(devmgr, mounts_tmpdir)
            for key, mount in self.mounts.items():
                data_mnt = mntmgr.mount(mount)
                mounts[key] = data_mnt

            self.prepare_arguments(args, args_path)

            api = API()
            build_root.register_api(api)

            rls = remoteloop.LoopServer()
            build_root.register_api(rls)

            extra_env = {}
            if self.source_epoch is not None:
                extra_env["SOURCE_DATE_EPOCH"] = str(self.source_epoch)
            if experimentalflags.get_bool("debug-qemu-user"):
                extra_env["QEMU_LOG"] = "unimp"

            debug_shell = debug_break in ('*', self.name, self.id)

            r = build_root.run([f"/run/osbuild/bin/{self.name}"],
                               monitor,
                               timeout=timeout,
                               binds=binds,
                               readonly_binds=ro_binds,
                               extra_env=extra_env,
                               debug_shell=debug_shell)

        return StageResult(self, r.returncode, r.output, api.error)


class Runner:
    def __init__(self, info, name: Optional[str] = None) -> None:
        self.info_path = info.path
        self.name = name or os.path.basename(info.path)

    @property
    def path(self):
        return self.info_path

    @property
    def exec(self):
        return os.path.basename(self.info_path)

    def to_dict(self, libdir):
        info_path_rel = Path(self.info_path).relative_to(libdir)
        return {
            "name": self.name,
            "info_path": str(info_path_rel),
        }

    @classmethod
    def from_dict(cls, data, libdir):
        info_path_abs = str(Path(libdir) / data["info_path"])

        class InfoStub:
            def __init__(self, path):
                self.path = path

        info = InfoStub(info_path_abs)
        return cls(info, data["name"])


# pylint: disable=too-many-branches
class Pipeline:
    def __init__(self, name: str, runner: Runner, build=None, source_epoch=None):
        self.name = name
        self.build = build
        self.runner = runner
        self.stages: List[Stage] = []
        self.assembler = None
        self.source_epoch = source_epoch
        self.run_in_vm = False

    @property
    def id(self):
        """
        Pipeline id: corresponds to the `id` of the last stage

        In contrast to `name` this identifies the pipeline via
        the tree, i.e. the content, it produces. Therefore two
        pipelines that produce the same `tree`, i.e. have the
        same exact stages and build pipeline, will have the
        same `id`; thus the `id`, in contrast to `name` does
        not uniquely identify a pipeline.
        In case a Pipeline has no stages, its `id` is `None`.
        """
        return self.stages[-1].id if self.stages else None

    def add_stage(self, info, options, sources_options=None):
        stage = Stage(info, sources_options, self.build,
                      self.id, options or {}, self.source_epoch)
        self.stages.append(stage)
        if self.assembler:
            self.assembler.base = stage.id
        return stage

    def build_stages(self, object_store, monitor, libdir,
                     debug_break="", stage_timeout=None, in_vm=False):
        results = []

        # If there are no stages, just return here
        if not self.stages:
            return results

        # Check if the tree that we are supposed to build does
        # already exist. If so, short-circuit here
        if object_store.contains(self.id):
            return results

        # We need a build tree for the stages below, which is either
        # another tree that needs to be built with the build pipeline
        # or the host file system if no build pipeline is specified
        # NB: the very last level of nested build pipelines is always
        # build on the host

        if not self.build:
            build_tree = object_store.host_tree
        else:
            build_tree = object_store.get(self.build)

        if not build_tree:
            raise AssertionError(f"build tree {self.build} not found")

        # Not in the store yet, need to actually build it, but maybe
        # an intermediate checkpoint exists: Find the last stage that
        # already exists in the store and use that as the base.
        tree = object_store.new(self.id)
        tree.source_epoch = self.source_epoch

        todo = collections.deque()
        for stage in reversed(self.stages):
            base = object_store.get(stage.id)
            if base:
                tree.init(base)
                break
            todo.append(stage)  # append right side of the deque

        # If two run() calls race each-other, two trees will get built
        # and it is nondeterministic which of them will end up
        # referenced by the `tree_id` in the content store if they are
        # both committed. However, after the call to commit all the
        # trees will be based on the winner.
        with contextlib.ExitStack() as cm:
            qemu = None
            tree_dir = os.path.join(tree.path, "tree")
            if in_vm:
                buildtree_dir = os.path.join(build_tree.path, "tree")

                qemu = Qemu("3G",
                            buildtree_dir,
                            libdir)
                qemu.add_virtiofs(object_store.store, "store", readonly=False)

                cm.enter_context(qemu)

                # Update the object store in the VM to match
                # the active floating objects
                qemu.monitored_request(monitor, "update_store",
                                       floating=object_store.export_floating())

            while todo:
                stage = todo.pop()

                monitor.stage(stage)

                with tree.meta.write(stage.id) as meta:
                    if qemu:
                        resp = qemu.monitored_request(monitor, "run_stage",
                                                      stage=stage.to_dict(libdir),
                                                      runner=self.runner.to_dict(libdir),
                                                      tree_dir=str(Path(tree_dir).relative_to(object_store.store)),
                                                      meta_name=str(Path(meta.name).relative_to(object_store.store)),
                                                      build_tree=build_tree.id,
                                                      debug_break=debug_break,
                                                      stage_timeout=stage_timeout)
                        r = StageResult.from_dict(resp)
                    else:
                        r = stage.run(os.fspath(tree),
                                      meta.name,
                                      self.runner,
                                      build_tree,
                                      object_store,
                                      monitor,
                                      libdir,
                                      debug_break,
                                      stage_timeout)

                md = tree.meta.get(r.id)
                monitor.result(r, md)
                results.append(r)
                if not r.success:
                    cleanup(build_tree, tree)
                    return results

                if stage.checkpoint:
                    object_store.commit(tree, stage.id)

        tree.finalize()

        return results

    def run(self, store, monitor, libdir, debug_break="", stage_timeout=None, in_vm=False):
        self.run_in_vm = in_vm
        monitor.begin(self)
        pr = PipelineResult(self.name)
        stage_results = self.build_stages(store,
                                          monitor,
                                          libdir,
                                          debug_break,
                                          stage_timeout,
                                          in_vm)
        pr.set_stages(stage_results)
        monitor.finish({"name": self.name})
        return pr


class DownloadResult:
    def __init__(self):
        self.success = True
        self.sources = {}

    def add(self, source_type, source_results: SourceResults):
        self.sources[source_type] = source_results
        if not source_results.success:
            self.success = False

    def as_dict(self):
        return {
            "success": self.success,
            "sources": {src_type: src.as_dict() for src_type, src in self.sources.items()}
        }


class PipelineResult:
    def __init__(self, name, success=True):
        self.success = success
        self.name = name
        self.stages = []

    def set_stages(self, stages: List[StageResult]):
        self.stages = stages
        self.success = all(sr.success for sr in stages)


class ManifestBuildResult:
    def __init__(self, dr):
        self.success = dr.success
        self.download_result = dr
        self.pipeline_results = {}

    def add_pipeline(self, pl_id: str, pr: PipelineResult):
        self.pipeline_results[pl_id] = pr
        if not pr.success:
            self.success = False


class Manifest:
    """Representation of a pipeline and its sources"""

    def __init__(self):
        self.metadata = {}
        self.pipelines = collections.OrderedDict()
        self.sources = []

    def add_metadata(self, name: str, data: Dict[str, Any]) -> None:
        self.metadata[name] = data

    def add_pipeline(
        self,
        name: str,
        runner: Runner,
        build: Optional[str] = None,
        source_epoch: Optional[int] = None
    ) -> Pipeline:
        pipeline = Pipeline(name, runner, build, source_epoch)
        if name in self.pipelines:
            raise ValueError(f"Name {name} already exists")
        self.pipelines[name] = pipeline
        return pipeline

    def add_source(self, info, items: List, options: Dict) -> Source:
        source = Source(info, items, options)
        self.sources.append(source)
        return source

    def download(self, store, monitor):
        dr = DownloadResult()
        with host.ServiceManager(monitor=monitor) as mgr:
            for source in self.sources:
                monitor.begin(source)
                result = source.download(mgr, store)
                monitor.result(result)
                dr.add(source.info_name, result)
                monitor.finish({"name": source.info_name})
        return dr

    def export_sources(self, store, dst_store, monitor):
        with host.ServiceManager(monitor=monitor) as mgr:
            for source in self.sources:
                monitor.begin(source)
                result = source.copy(mgr, store, dst_store)
                monitor.result(result)
                monitor.finish({"name": source.info_name})

    def depsolve(self, store: ObjectStore, targets: Iterable[str]) -> List[str]:
        """Return the list of pipelines that need to be built

        Given a list of target pipelines, return the names
        of all pipelines and their dependencies that are not
        already present in the store.
        """

        # A stack of pipelines to check if they need to be built
        check = list(map(self.get, targets))

        # The ordered result "set", will be reversed at the end
        build = collections.OrderedDict()

        while check:
            pl = check.pop()  # get the last(!) item

            if not pl:
                raise RuntimeError("Could not find pipeline.")

            if store.contains(pl.id):
                continue

            # The store does not have this pipeline, it needs to
            # be built, add it to the ordered result set and
            # ensure it is at the end, i.e. built before previously
            # checked items. NB: the result set is reversed before
            # it gets returned. This ensures that a dependency that
            # gets checked multiple times, like a build pipeline,
            # always gets built before its dependent pipeline.
            build[pl.id] = pl
            build.move_to_end(pl.id)

            # Add all dependencies to the stack of things to check,
            # starting with the build pipeline, if there is one
            if pl.build:
                check.append(self.get(pl.build))

            # Stages depend on other pipeline via pipeline inputs.
            # We check in reversed order until we hit a checkpoint
            for stage in reversed(pl.stages):

                # we stop if we have a checkpoint, i.e. we don't
                # need to build any stages after that checkpoint
                if store.contains(stage.id):
                    break

                pls = map(self.get, stage.dependencies)
                check.extend(pls)

        return list(map(lambda x: x.name, reversed(build.values())))

    def build(self, store, pipelines, monitor, libdir,
              debug_break="", stage_timeout=None, in_vm=None) -> ManifestBuildResult:
        """Build the manifest

        Returns a dict of string keys that contains the overall
        "success" and the `StageResult` of each individual pipeline.

        The overall success "success" is stored as the string "success"
        with the bool result and the build pipelines BuildStatus is
        stored under the pipelines ID string.
        """
        dr = self.download(store, monitor)
        mbr = ManifestBuildResult(dr)
        if not mbr.success:
            return mbr

        for name_or_id in pipelines:
            pl = self[name_or_id]
            res = pl.run(store, monitor, libdir, debug_break, stage_timeout, in_vm=in_vm and pl.name in in_vm)
            mbr.add_pipeline(pl.id, res)
            if not mbr.success:
                return mbr

        return mbr

    def mark_checkpoints(self, patterns):
        """Match pipeline names, stage ids, and stage names against an iterable
        of `fnmatch`-patterns."""
        selected = set()

        def matching(haystack):
            return any(fnmatch(haystack, p) for p in patterns)

        for pipeline in self.pipelines.values():
            # checkpoints are marked on stages, if a pipeline has no stages we
            # can't mark it
            if not pipeline.stages:
                continue

            if matching(pipeline.name):
                selected.add(pipeline.id)
                pipeline.stages[-1].checkpoint = True

            for stage in pipeline.stages:
                if matching(stage.id) or matching(stage.name):
                    selected.add(stage.id)
                    stage.checkpoint = True

        return selected

    def get(self, name_or_id: str) -> Optional[Pipeline]:
        pl = self.pipelines.get(name_or_id)
        if pl:
            return pl
        for pl in self.pipelines.values():
            if pl.id == name_or_id:
                return pl
        return None

    def __contains__(self, name_or_id: str) -> bool:
        return self.get(name_or_id) is not None

    def __getitem__(self, name_or_id: str) -> Pipeline:
        pl = self.get(name_or_id)
        if pl:
            return pl
        raise KeyError(f"'{name_or_id}' not found in manifest pipelines: {list(self.pipelines.keys())}")

    def __iter__(self) -> Iterator[Pipeline]:
        return iter(self.pipelines.values())


def detect_host_runner():
    """Use os-release(5) to detect the runner for the host"""
    osname = osrelease.describe_os(*osrelease.DEFAULT_PATHS)
    return "org.osbuild." + osname



import os
import pathlib
from types import ModuleType

import pytest

import osbuild.meta
from osbuild import sources, testutil
from osbuild.testutil.imports import import_module_from_path


@pytest.fixture(name="sources_module")
def sources_module_fixture(request: pytest.FixtureRequest) -> ModuleType:
    """sources_module is a fixture that imports a stage module by its name
    defined in SOURCES_NAME in the test module.
    """
    if not hasattr(request.module, "SOURCES_NAME"):
        raise ValueError("sources_module fixture must be used in a test module that defines SOURCES_NAME")

    sources_name = request.module.SOURCES_NAME
    caller_dir = pathlib.Path(request.node.fspath).parent
    module_path = caller_dir.parent / sources_name
    return import_module_from_path("sources", os.fspath(module_path))


@pytest.fixture
def sources_service(sources_module) -> ModuleType:
    """sources_service is a fixture that imports a sources module by its name
    defined in SOURCES_NAME in the test module and returns a SourcesService
    """
    service_cls = testutil.find_one_subclass_in_module(sources_module, sources.SourceService)
    fd = testutil.make_fake_service_fd()
    srv_obj = service_cls.from_args(["--service-fd", str(fd)])
    return srv_obj


@pytest.fixture
def sources_schema(request: pytest.FixtureRequest) -> osbuild.meta.Schema:
    """
    source_schema is a fixture returns the schema for a sources module.
    """
    if hasattr(request, "param") and not isinstance(request.param, str):
        raise ValueError(
            "sources_schema fixture may be indirectly parametrized only with the sources schema version string")

    if not hasattr(request.module, "SOURCES_NAME"):
        raise ValueError("sources_schema fixture must be used in a test module that defines SOURCES_NAME")

    sources_name = request.module.SOURCES_NAME
    schema_version = request.param if hasattr(request, "param") else "2"
    caller_dir = pathlib.Path(request.node.fspath).parent
    root = caller_dir.parent.parent
    mod_info = osbuild.meta.ModuleInfo.load(root, "Source", sources_name)
    return osbuild.meta.Schema(mod_info.get_schema(version=schema_version), sources_name)



"""
Utility functions for reading and writing toml files.

Handles module imports for all supported versions (in a build root or on a host).
"""
import importlib
from types import ModuleType
from typing import Optional

# Different modules require different file mode (text vs binary)
_toml_modules = {
    "tomllib": {"mode": "rb"},  # stdlib since 3.11 (read-only)
    "tomli": {"mode": "rb"},  # EL9+
    "toml": {"mode": "r", "encoding": "utf-8"},  # older unmaintained lib, needed for backwards compatibility
    "pytoml": {"mode": "r", "encoding": "utf-8"},  # deprecated, needed for backwards compatibility (EL8 manifests)
}


_toml: Optional[ModuleType] = None
_rargs: dict = {}
for module, args in _toml_modules.items():
    try:
        _toml = importlib.import_module(module)
        _rargs = args
        break
    except ModuleNotFoundError:
        pass
else:
    raise ModuleNotFoundError("No toml module found: " + ", ".join(_toml_modules))

# Different modules require different file mode (text vs binary)
_tomlw_modules = {
    "tomli_w": {"mode": "wb"},  # EL9+
    "toml": {"mode": "w", "encoding": "utf-8"},  # older unmaintained lib, needed for backwards compatibility
    "pytoml": {"mode": "w", "encoding": "utf-8"},  # deprecated, needed for backwards compatibility (EL8 manifests)
}


_tomlw: Optional[ModuleType] = None
_wargs: dict = {}
for module, args in _tomlw_modules.items():
    try:
        _tomlw = importlib.import_module(module)
        _wargs = args
        break
    except ModuleNotFoundError:
        # allow importing without write support
        pass


def load_from_file(path):
    if _toml is None:
        raise RuntimeError("no toml module available")

    with open(path, **_rargs) as tomlfile:  # pylint: disable=unspecified-encoding
        return _toml.load(tomlfile)


def dump_to_file(data, path, header=""):
    if _tomlw is None:
        raise RuntimeError("no toml module available with write support")

    with open(path, **_wargs) as tomlfile:  # pylint: disable=unspecified-encoding
        if header:
            _write_comment(tomlfile, header)

        _tomlw.dump(data, tomlfile)


def _write_comment(f, comment: list):
    if not comment:
        return

    data = "\n".join(map(lambda c: f"# {c}", comment)) + "\n\n"
    if "b" in f.mode:
        f.write(data.encode())
    else:
        f.write(data)



#!/usr/bin/python3

import os.path
import subprocess
import uuid
from unittest import mock

import pytest

from osbuild import testutil
from osbuild.testutil import has_executable

STAGE_NAME = "org.osbuild.mkfs.xfs"


# Prepare dataset containing good and bad API call parameters
@pytest.mark.parametrize("test_data, expected_err", [
    # Bad API parameters
    ({}, "'uuid' is a required property"),
    ({"uuid": 123}, "123 is not of type 'string'"),
    ({"uuid": "text", "label": 1234}, "1234 is not of type 'string'"),
    ({"uuid": "text", "label": "12345678901234567"}, "is too long"),
    # Good API parameters
    # mkfs requires valid UUID, but API will accept any string right now.
    # Bellow will fail when API will be more restricted in the future.
    ({"uuid": "text"}, ""),
    ({"uuid": "text", "label": "123456789012"}, ""),
])
# This test validates only API calls using correct and incorrect queries
def test_schema_validation_mkfs_xfs(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "devices": {
            "device": {
                "path": "some-path",
            },
        },
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


# This test creates dummy filesystem using predefined parameters and verifies
# that end result has those parameters set
@pytest.mark.skipif(not has_executable("mkfs.xfs"), reason="need mkfs.xfs")
def test_mkfs_xfs_integration(tmp_path, stage_module):
    fake_disk_path = tmp_path / "fake.img"
    with fake_disk_path.open("w") as fp:
        fp.truncate(300 * 1024 * 1024)
    devices = {
        "device": {
            "path": fake_disk_path,
        },
    }
    fake_uuid = str(uuid.uuid4())
    fake_label = "osbuild"
    options = {
        "uuid": fake_uuid,
        "label": fake_label,
    }
    stage_module.main(devices, options)
    assert os.path.exists(fake_disk_path)
    output = subprocess.check_output([
        "xfs_admin", "-l", "-u", fake_disk_path], encoding="utf-8")
    assert f'UUID = {fake_uuid}' in output, \
        f'expected UUID not found in: {output}'
    assert f'label = "{fake_label}"' in output, \
        f'expected label not found in: {output}'


@pytest.mark.parametrize("test_input, expected", [
    ({}, []),
    ({"label": "osbuild"}, ["-L", "osbuild"]),
])
@mock.patch("subprocess.run")
def test_mkfs_xfs_cmdline(mock_run, stage_module, test_input, expected):
    fake_disk_path = "/dev/xxd1"
    devices = {
        "device": {
            "path": fake_disk_path,
        },
    }
    fake_uuid = str(uuid.uuid4())
    options = {
        "uuid": fake_uuid,
    }
    options.update(test_input)
    stage_module.main(devices, options)

    expected = [
        "mkfs.xfs",
        "-m",
        f"uuid={fake_uuid}",
    ] + expected + [
        fake_disk_path,
    ]
    mock_run.assert_called_once_with(expected, encoding="utf8", check=True)



#!/usr/bin/python3

import re

import pytest

from osbuild.testutil import (
    assert_jsonschema_error_contains,
)

STAGE_NAME = "org.osbuild.waagent.conf"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({}, r"'config' is a required property"),
    ({"config": {"Unsupported": True}}, r"Additional properties are not allowed \('Unsupported' was unexpected\)"),
    ({"config": {"Provisioning.UseCloudInit": 42}}, r"42 is not of type 'boolean'"),
    ({"config": {"Provisioning.Enabled": "abc"}}, r"'abc' is not of type 'boolean'"),
    ({"config": {"ResourceDisk.Format": 42}}, r"42 is not of type 'boolean'"),
    ({"config": {"ResourceDisk.EnableSwap": "abc"}}, r"'abc' is not of type 'boolean'"),
    # good
    ({"config": {}}, ""),
    ({
        "config": {
            "Provisioning.UseCloudInit": True,
            "Provisioning.Enabled": True,
            "ResourceDisk.Format": False,
            "ResourceDisk.EnableSwap": False,
        },
    }, ""),
])
def test_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {},
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        assert_jsonschema_error_contains(res, re.compile(expected_err), expected_num_errs=1)



#!/usr/bin/python3

import os
from unittest import mock

import pytest

from osbuild.testutil import make_fake_input_tree

STAGE_NAME = "org.osbuild.xorrisofs"


@mock.patch("subprocess.run")
def test_xorrisofs_syslinux(mock_run, tmp_path, stage_module):

    fake_input_tree = make_fake_input_tree(tmp_path, {})

    inputs = {
        "tree": {
            "path": fake_input_tree,
        }
    }

    # default (syslinux)
    stage_module.main(
        inputs,
        tmp_path,
        {
            "boot": {
                "image": "images/eltorito.img",
                "catalog": "boot.cat",
            },
            "filename": "test.iso",
            "volid": "test",
            "efi": "images/efiboot.img",
        },
    )

    mock_run.assert_called_with([
        "/usr/bin/xorrisofs",
        "-verbose",
        "-V", "test",
        "-b", "images/eltorito.img",
        "-c", "boot.cat",
        "--boot-catalog-hide",
        "-boot-load-size", "4",
        "-boot-info-table",
        "-no-emul-boot",
        "-rock",
        "-joliet",
        "-eltorito-alt-boot",
        "-e", "images/efiboot.img",
        "-no-emul-boot",
        "-o", os.path.join(tmp_path, "test.iso"),
        fake_input_tree,
    ], check=True)


@mock.patch("subprocess.run")
def test_xorrisofs_syslinux_with_excludes(mock_run, tmp_path, stage_module):
    fake_input_tree = make_fake_input_tree(tmp_path, {})

    inputs = {
        "tree": {
            "path": fake_input_tree,
        }
    }

    # default (syslinux)
    stage_module.main(
        inputs,
        tmp_path,
        {
            "boot": {
                "image": "images/eltorito.img",
                "catalog": "boot.cat",
            },
            "filename": "test.iso",
            "volid": "test",
            "efi": "images/efiboot.img",
            "exclude": [
                "efiboot.img",
            ],
        },
    )

    mock_run.assert_called_with([
        "/usr/bin/xorrisofs",
        "-verbose",
        "-V", "test",
        "-b", "images/eltorito.img",
        "-c", "boot.cat",
        "--boot-catalog-hide",
        "-boot-load-size", "4",
        "-boot-info-table",
        "-no-emul-boot",
        "-rock",
        "-joliet",
        "-eltorito-alt-boot",
        "-e", "images/efiboot.img",
        "-no-emul-boot",
        "-exclude", "efiboot.img",
        "-o", os.path.join(tmp_path, "test.iso"),
        fake_input_tree,
    ], check=True)


@mock.patch("subprocess.run")
def test_xorrisofs_syslinux_nobootinfo(mock_run, tmp_path, stage_module):

    fake_input_tree = make_fake_input_tree(tmp_path, {})

    inputs = {
        "tree": {
            "path": fake_input_tree,
        }
    }

    # default (syslinux)
    stage_module.main(
        inputs,
        tmp_path,
        {
            "boot": {
                "image": "images/eltorito.img",
                "catalog": "boot.cat",
                "nobootinfo": True,
            },
            "filename": "test.iso",
            "volid": "test",
            "efi": "images/efiboot.img",
        },
    )

    mock_run.assert_called_with([
        "/usr/bin/xorrisofs",
        "-verbose",
        "-V", "test",
        "-b", "images/eltorito.img",
        "-c", "boot.cat",
        "--boot-catalog-hide",
        "-boot-load-size", "4",
        "-no-emul-boot",
        "-rock",
        "-joliet",
        "-eltorito-alt-boot",
        "-e", "images/efiboot.img",
        "-no-emul-boot",
        "-o", os.path.join(tmp_path, "test.iso"),
        fake_input_tree,
    ], check=True)


@mock.patch("subprocess.run")
@mock.patch("os.path.exists")
def test_xorrisofs_grub2(mock_exists, mock_run, tmp_path, stage_module):
    mock_exists.return_value = True

    fake_input_tree = make_fake_input_tree(tmp_path, {})

    inputs = {
        "tree": {
            "path": fake_input_tree,
        }
    }

    stage_module.main(
        inputs,
        tmp_path,
        {
            "boot": {
                "image": "images/eltorito.img",
                "catalog": "boot.cat",
            },
            "grub2mbr": "/usr/lib/grub/i386-pc/boot_hybrid.img",
            "filename": "test.iso",
            "volid": "test",
            "efi": "images/efiboot.img",
        },
    )

    mock_run.assert_called_with([
        "/usr/bin/xorrisofs",
        "-verbose",
        "-rock",
        "-joliet",
        "-V", "test",
        "--grub2-mbr", "/usr/lib/grub/i386-pc/boot_hybrid.img",
        "-partition_offset", "16",
        "-appended_part_as_gpt",
        "-append_partition", "2", "C12A7328-F81F-11D2-BA4B-00A0C93EC93B", os.path.join(fake_input_tree, "images/efiboot.img"),
        "-iso_mbr_part_type", "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "-b", "images/eltorito.img",
        "-c", "boot.cat",
        "--boot-catalog-hide",
        "-no-emul-boot",
        "-boot-load-size", "4",
        "-boot-info-table",
        "--grub2-boot-info",
        "-eltorito-alt-boot",
        "-e", "--interval:appended_partition_2:all::",
        "-no-emul-boot",
        "-o", os.path.join(tmp_path, "test.iso"),
        fake_input_tree,
    ], check=True)


@mock.patch("subprocess.run")
@mock.patch("os.path.exists")
def test_xorrisofs_grub2_with_excludes(mock_exists, mock_run, tmp_path, stage_module):
    mock_exists.return_value = True

    fake_input_tree = make_fake_input_tree(tmp_path, {})

    inputs = {
        "tree": {
            "path": fake_input_tree,
        }
    }

    stage_module.main(
        inputs,
        tmp_path,
        {
            "boot": {
                "image": "images/eltorito.img",
                "catalog": "boot.cat",
            },
            "grub2mbr": "/usr/lib/grub/i386-pc/boot_hybrid.img",
            "filename": "test.iso",
            "volid": "test",
            "efi": "images/efiboot.img",
            "exclude": [
                "efiboot.img",
            ],
        },
    )

    mock_run.assert_called_with([
        "/usr/bin/xorrisofs",
        "-verbose",
        "-rock",
        "-joliet",
        "-V", "test",
        "--grub2-mbr", "/usr/lib/grub/i386-pc/boot_hybrid.img",
        "-partition_offset", "16",
        "-appended_part_as_gpt",
        "-append_partition", "2", "C12A7328-F81F-11D2-BA4B-00A0C93EC93B", os.path.join(fake_input_tree, "images/efiboot.img"),
        "-iso_mbr_part_type", "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "-b", "images/eltorito.img",
        "-c", "boot.cat",
        "--boot-catalog-hide",
        "-no-emul-boot",
        "-boot-load-size", "4",
        "-boot-info-table",
        "--grub2-boot-info",
        "-eltorito-alt-boot",
        "-e", "--interval:appended_partition_2:all::",
        "-no-emul-boot",
        "-exclude", "efiboot.img",
        "-o", os.path.join(tmp_path, "test.iso"),
        fake_input_tree,
    ], check=True)


@mock.patch("subprocess.run")
@mock.patch("os.path.exists")
def test_xorrisofs_grub2_ppc64le(mock_exists, mock_run, tmp_path, stage_module):
    mock_exists.return_value = True

    fake_input_tree = make_fake_input_tree(tmp_path, {})

    inputs = {
        "tree": {
            "path": fake_input_tree,
        }
    }

    stage_module.main(
        inputs,
        tmp_path,
        {
            "filename": "test.iso",
            "volid": "test",
            "prep": "image-builder",
            "rational_rock": True,
            "untranslated_filenames": True,
            "full_filenames": True,
            "volset": "Fedora-43-Everything-ppc64le",
            "chrp_boot": True,
        },
    )

    mock_run.assert_called_with([
        "/usr/bin/xorrisofs",
        "-verbose",
        "-rock",
        "-joliet",
        "-V", "test",
        "-preparer", "image-builder",
        "-rational-rock",
        "-untranslated-filenames",
        "-full-iso9660-filenames",
        "-volset", "Fedora-43-Everything-ppc64le",
        "-volset-size", "1",
        "-volset-seqno", "1",
        "-chrp-boot",
        "-o", os.path.join(tmp_path, "test.iso"),
        fake_input_tree,
    ], check=True)


# NOTE: run is mocked so that if an error isn't raised it won't run xorrisofs
@mock.patch("subprocess.run")
def test_xorrisofs_grub2_ppc64le_efi_error(_mock_run, tmp_path, stage_module):
    fake_input_tree = make_fake_input_tree(tmp_path, {})

    inputs = {
        "tree": {
            "path": fake_input_tree,
        }
    }

    with pytest.raises(RuntimeError):
        stage_module.main(
            inputs,
            tmp_path,
            {
                "filename": "test.iso",
                "volid": "test",
                "efi": "images/efiboot.img",
                "chrp_boot": True,
            },
        )



#!/usr/bin/python3

import os.path

STAGE_NAME = "org.osbuild.cloud-init"


# Test that the 'datasource_list' option is dumped as a single line
# in the configuration file.
def test_cloud_init_datasource_list_single_line(tmp_path, stage_module):
    treedir = tmp_path / "tree"
    confpath = treedir / "etc/cloud/cloud.cfg.d/datasource_list.cfg"
    confpath.parent.mkdir(parents=True, exist_ok=True)

    options = {
        "filename": confpath.name,
        "config": {
            "datasource_list": [
                "Azure",
            ]
        }
    }

    stage_module.main(treedir, options)
    assert os.path.exists(confpath)
    assert confpath.read_text() == "datasource_list: [Azure]\n"


def test_cloud_init_network(tmp_path, stage_module):
    treedir = tmp_path / "tree"
    confpath = treedir / "etc/cloud/cloud.cfg.d/network.cfg"
    confpath.parent.mkdir(parents=True, exist_ok=True)

    options = {
        "filename": confpath.name,
        "config": {
            "network": {
                "config": "disabled",
            },
        },
    }

    stage_module.main(treedir, options)
    assert os.path.exists(confpath)
    assert confpath.read_text() == "network:\n  config: disabled\n"



#!/usr/bin/python3

from unittest.mock import call, patch

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.bootupd"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({"deployment": "must-be-object"}, "'must-be-object' is not of type 'object'"),
    ({"deployment": {"osname": "some-os"}}, "{'osname': 'some-os'} is not valid under any of the given schemas"),
    ({"deployment": {"ref": "some-ref"}}, "{'ref': 'some-ref'} is not valid under any of the given schemas"),
    ({"deployment": {"osname": "some-os", "ref": "some-ref", "serial": "must-be-number"}},
     "'must-be-number' is not of type 'number'"),
    ({"deployment": {"default": False}}, "{'default': False} is not valid under any of the given schemas"),
    ({"deployment": {
        "osname": "some-os",
        "ref": "some-ref",
        "serial": 0,
        "default": True}
      }, "{'osname': 'some-os', 'ref': 'some-ref', 'serial': 0, 'default': True} is not valid under any of the given schemas"),
    ({"random": "property"}, "Additional properties are not allowed"),
    ({"bios": {}}, "'device' is a required property"),
    ({"bios": "must-be-object"}, "'must-be-object' is not of type 'object'"),
    # good
    ({}, ""),
    ({
        "deployment":
        {
            "osname": "some-os",
            "ref": "some-ref",
            "serial": 1,
        },
        "static-configs": True,
        "bios":
        {
            "device": "/dev/sda",
        },
    }, ""),
    ({
        "deployment":
        {
            "default": True,
        },
        "static-configs": True,
        "bios":
        {
            "device": "/dev/sda",
        },
    }, "")

])
def test_bootupd_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


@pytest.mark.parametrize("test_options,expected_bootupd_opts", [
    ({}, []),
    ({"bios": {"device": "/dev/sda"}}, ["--device=/mapped/dev/sda"]),
    ({"static-configs": True}, ["--with-static-configs"]),
])
@patch("subprocess.run")
def test_bootupd_defaults(mocked_run, stage_module, test_options, expected_bootupd_opts):
    args = {
        "paths": {
            "mounts": "/run/osbuild/mounts",
        },
        "devices": {
            "/dev/sda": {
                "path": "/mapped/dev/sda",
            },
        },
    }
    options = test_options
    with patch.object(stage_module, "bind_mounts") as mocked_bind_mounts:
        stage_module.main(args, options)

    assert len(mocked_bind_mounts.call_args_list) == 1
    assert mocked_run.call_args_list == [call(["chroot",
                                               "/run/osbuild/mounts",
                                               "/usr/bin/bootupctl",
                                               "backend",
                                               "install"] + expected_bootupd_opts + ["/run/osbuild/mounts"],
                                              check=True),
                                         ]


@patch("subprocess.run")
def test_bootupd_bind_mounts(mocked_run, stage_module):
    dst = "/run/osbuild/mounts/"
    with stage_module.bind_mounts(['/dev', '/proc', '/sys', '/run', '/var', '/tmp'], dst):
        pass
    assert mocked_run.call_args_list == [
        call(["mount", "--rbind", "/dev", "/run/osbuild/mounts/dev"], check=True),
        call(["mount", "--rbind", "/proc", "/run/osbuild/mounts/proc"], check=True),
        call(["mount", "--rbind", "/sys", "/run/osbuild/mounts/sys"], check=True),
        call(["mount", "--rbind", "/run", "/run/osbuild/mounts/run"], check=True),
        call(["mount", "--rbind", "/var", "/run/osbuild/mounts/var"], check=True),
        call(["mount", "--rbind", "/tmp", "/run/osbuild/mounts/tmp"], check=True),
        # and umount in reverse order
        call(["umount", "--recursive", "/run/osbuild/mounts/tmp"], check=False),
        call(["umount", "--recursive", "/run/osbuild/mounts/var"], check=False),
        call(["umount", "--recursive", "/run/osbuild/mounts/run"], check=False),
        call(["umount", "--recursive", "/run/osbuild/mounts/sys"], check=False),
        call(["umount", "--recursive", "/run/osbuild/mounts/proc"], check=False),
        call(["umount", "--recursive", "/run/osbuild/mounts/dev"], check=False),
    ]



#!/usr/bin/python3
"""
Utility functions to inspect PE32+ (Portable Executable) files

To read all the section headers of an PE32+ file[1], while also
inspecting the individual headers, the `coff` header can be passed
to the individual function, which avoids having to re-read it:

```
with open("file.pe", "rb") as f:
    coff = pe32p.read_coff_header(f)
    opt = pe32p.read_optional_header(f, coff)
    sections = pe32p.read_sections(f, coff)
```

Passing `coff` to the functions eliminates extra i/o to seek to the correct
file positions, but it requires that the functions are called in the given
order, i.e. `read_coff_header`, `read_optional_haeder` then `read_sections`.

[1] https://learn.microsoft.com/en-us/windows/win32/debug/pe-format
"""

import enum
import io
import os
import struct
import sys
from collections import namedtuple
from typing import BinaryIO, Iterator, List, Optional, Union

PathLike = Union[str, bytes, os.PathLike]

CoffFormat = "4sHHIIIHH"
CoffHeader = namedtuple(
    "CoffHeader",
    [
        "Signature",
        "Machine",
        "NumberOfSections",
        "TimeDateStamp",
        "PointerToSymbolTable",
        "NumberOfSymbols",
        "SizeOfOptionalHeader",
        "Characteristics",
    ]
)


SectionFormat = "8sIIIIIIHHI"
SectionHeader = namedtuple(
    "SectionHeader",
    [
        "Name",
        "VirtualSize",
        "VirtualAddress",
        "SizeOfRawData",
        "PointerToRawData",
        "PointerToRelocations",
        "PointerToLinenumbers",
        "NumberOfRelocations",
        "NumberOfLinenumbers",
        "Characteristics",
    ]
)


class SectionFlags(enum.Flag):
    ALIGN_1BYTES = 0x00100000
    ALIGN_2BYTES = 0x00200000
    ALIGN_4BYTES = 0x00300000
    ALIGN_8BYTES = 0x00400000
    ALIGN_16BYTES = 0x00500000
    ALIGN_32BYTES = 0x00600000
    ALIGN_64BYTES = 0x00700000
    ALIGN_128BYTES = 0x00800000
    ALIGN_256BYTES = 0x00900000
    ALIGN_512BYTES = 0x00A00000
    ALIGN_1024BYTES = 0x00B00000
    ALIGN_2048BYTES = 0x00C00000
    ALIGN_4096BYTES = 0x00D00000
    ALIGN_8192BYTES = 0x00E00000
    ALIGN_MASK = 0x00F00000
    ALIGN_DEFAULT = ALIGN_16BYTES


OptionalFormat = "HBBIIIIIQIIHHHHHHIIIIHHQQQQII"
OptionalHeader = namedtuple(
    "OptionalHeader",
    [
        # Standard fields
        "Magic",
        "MajorLinkerVersion",
        "MinorLinkerVersion",
        "SizeOfCode",
        "SizeOfInitializedData",
        "SizeOfUninitializedData",
        "AddressOfEntryPoint",
        "BaseOfCode",
        # Windows-Specific fields (PE32+)
        "ImageBase",
        "SectionAlignment",
        "FileAlignment",
        "MajorOperatingSystemVersion",
        "MinorOperatingSystemVersion",
        "MajorImageVersion",
        "MinorImageVersion",
        "MajorSubsystemVersion",
        "MinorSubsystemVersion",
        "Reserved1",
        "SizeOfImage",
        "SizeOfHeaders",
        "CheckSum",
        "Subsystem",
        "DllCharacteristics",
        "SizeOfStackReserve",
        "SizeOfStackCommit",
        "SizeOfHeapReserve",
        "SizeOfHeapCommit",
        "LoaderFlags",
        "NumberOfRvaAndSizes",
    ]
)


def read_coff_header(f: BinaryIO) -> CoffHeader:
    """Read the Common Object File Format (COFF) Header of the open file at `f`"""

    # Quote from the "PE Format" article (see [1] in this module's doc string):
    # "[...] at the file offset specified at offset 0x3c, is a 4-byte signature
    # that identifies the file as a PE format image file. This signature is
    # 'PE\0\0' (the letters "P" and "E" followed by two null bytes). [...]
    # immediately after the signature of an image file, is a standard COFF
    # file header in the following format."
    # Our `CoffHeader` embeds the signature inside the CoffHeader.

    f.seek(0x3c, io.SEEK_SET)
    buf = f.read(struct.calcsize("I"))
    (s, ) = struct.unpack_from("I", buf)
    f.seek(int(s), io.SEEK_SET)

    buf = f.read(struct.calcsize(CoffFormat))
    coff = CoffHeader._make(struct.unpack_from(CoffFormat, buf))
    assert coff.Signature == b"PE\0\0", "Not a PE32+ file (missing PE header)"
    return coff


def read_optional_header(f: BinaryIO, coff: Optional[CoffHeader] = None) -> OptionalHeader:
    """Read the optional header of the open file at `f`

    If `coff` is passed in, the file position must point to directly after the
    COFF header, i.e. as if `read_coff_header` was just called.
    """
    if coff is None:
        coff = read_coff_header(f)

    buf = f.read(coff.SizeOfOptionalHeader)
    sz = struct.calcsize(OptionalFormat)
    assert len(buf) >= sz, "Optional header too small"
    opt = OptionalHeader._make(struct.unpack_from(OptionalFormat, buf))
    assert opt.Magic == 0x20B, f"Not a PE32+ file (magic: {opt.Magic:X})"
    return opt


def iter_sections(f: BinaryIO, coff: Optional[CoffHeader] = None) -> Iterator[SectionHeader]:
    """Iterate over all the sections in the open file at `f`

    If `coeff` is passed in, the file position must point directly after the Optional
    Header, i.e. as if `read_optional_haeder` was just called."""
    if coff is None:
        coff = read_coff_header(f)
        f.seek(coff.SizeOfOptionalHeader, io.SEEK_CUR)

    for _ in range(coff.NumberOfSections):
        buf = f.read(struct.calcsize(SectionFormat))
        yield SectionHeader._make(struct.unpack_from(SectionFormat, buf))


def read_sections(f: BinaryIO, coff: Optional[CoffHeader] = None) -> List[SectionHeader]:
    """Read all sections of the open file at `f`

    Like `iter_sections` but returns a list of `SectionHeader` objects."""
    return list(iter_sections(f, coff))


def main():

    if len(sys.argv) != 2:
        print(f"usage: {sys.argv[0]} FILE")
        sys.exit(1)

    with open(sys.argv[1], "rb") as f:
        coff = read_coff_header(f)
        opt = read_optional_header(f, coff)
        sections = read_sections(f, coff)

    print(coff)
    print(opt)
    for s in sections:
        print(s)

    last = sections[-1]
    print(f"{last.VirtualAddress: X}, {last.VirtualSize:X}")


if __name__ == "__main__":
    main()



import os.path
import re
from unittest import mock

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.hmac"


@pytest.mark.parametrize("test_data,expected_errs", [
    # bad
    ({}, [r"'paths' is a required property", r"'algorithm' is a required property"]),
    ({"algorithm": "sha512"}, [r"'paths' is a required property"]),
    ({"paths": ["/somefile"]}, [r"'algorithm' is a required property"]),
    ({"paths": [], "algorithm": "sha1"}, [r"(\[\] is too short|\[\] should be non-empty)"]),

    ({"paths": ["/somefiles"], "algorithm": "md5"},
     [r"'md5' is not one of \['sha1', 'sha224', 'sha256', 'sha384', 'sha512'\]"]),

    # good
    (
        {
            "paths": [
                "/greet1.txt",
                "/greet2.txt"
            ],
            "algorithm": "sha256"
        },
        ""
    ),
    (
        {
            "paths": [
                "/path/to/some/file",
                "/boot/efi/EFI/Linux/vmlinuz-linux"
            ],
            "algorithm": "sha512"
        },
        ""
    ),
])
def test_schema_validation_hmac(stage_schema, test_data, expected_errs):
    test_input = {
        "type": STAGE_NAME,
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if not expected_errs:
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        for exp_err in expected_errs:
            testutil.assert_jsonschema_error_contains(res, re.compile(exp_err), expected_num_errs=len(expected_errs))


@mock.patch("subprocess.run")
@pytest.mark.parametrize("options", [
    ({"paths": ["/a/file/path"], "algorithm": "sha512"}),
    ({"paths": ["/b/file/path"], "algorithm": "sha512"}),
    ({"paths": ["/a/file/path", "/b/file/path"], "algorithm": "sha256"}),
])
def test_hmac_cmdline(mock_run, tmp_path, stage_module, options):
    algorithm = options["algorithm"]

    expected_cmds = []
    expected_wds = []
    hmac_paths = []
    for path in options["paths"]:
        # create parent directories because the stage opens a file to write the output to it even when mocking sp.run()
        basename = os.path.basename(path)
        real_path = os.path.join(tmp_path, path.lstrip("/"))
        parent = os.path.dirname(real_path)
        hmac_paths.append(os.path.join(parent, f".{basename}.hmac"))
        os.makedirs(parent, exist_ok=True)
        expected_cmds.append([f"{algorithm}hmac", basename])
        expected_wds.append(parent)

    stage_module.main(tmp_path, options)
    for exp, cwd, actual in zip(expected_cmds, expected_wds, mock_run.call_args_list):
        assert exp == actual[0][0]
        assert cwd == actual[1]["cwd"]

    # the file should have been created by the open() call, but should be empty because we mocked sp.run()
    for path in hmac_paths:
        info = os.stat(path)
        assert info.st_size == 0



#!/usr/bin/python3

import os
from typing import List
from unittest.mock import patch

try:
    import librepo
except ImportError:
    import pytest
    pytest.skip("need librepo to run this test", allow_module_level=True)
import pytest

from osbuild import testutil
from osbuild.testutil.net import http_serve_directory

SOURCES_NAME = "org.osbuild.librepo"


@patch("librepo.download_packages")
def test_librepo_download_mocked(mocked_download_pkgs, sources_service):
    TEST_SOURCES = {
        "sha256:1111111111111111111111111111111111111111111111111111111111111111": {
            "path": "Packages/a/a",
            "mirror": "mirror_id",
        },
        "sha256:2111111111111111111111111111111111111111111111111111111111111111": {
            "path": "Packages/b/b",
            "mirror": "mirror_id2",
        },
    }
    sources_service.options = {
        "mirrors": {
            "mirror_id": {
                "url": "http://example.com/mirrorlist",
                "type": "mirrorlist",
            },
            "mirror_id2": {
                "url": "http://example.com/mirrorlist2",
                "type": "mirrorlist",
            }
        }
    }
    sources_service.cache = "cachedir"
    sources_service.fetch_all(TEST_SOURCES)
    # we expect one download call per mirror
    assert len(mocked_download_pkgs.call_args_list) == 2
    # extract the list of packages to download (all_calls->call()->args->args[0])
    download_pkgs = mocked_download_pkgs.call_args_list[0][0][0]
    assert download_pkgs[0].checksum == "1111111111111111111111111111111111111111111111111111111111111111"
    assert download_pkgs[0].checksum_type == librepo.SHA256
    assert download_pkgs[0].relative_url == "Packages/a/a"
    assert download_pkgs[0].dest == "cachedir/sha256:1111111111111111111111111111111111111111111111111111111111111111"
    assert download_pkgs[0].handle.mirrorlisturl == "http://example.com/mirrorlist"
    # and now the second call
    download_pkgs = mocked_download_pkgs.call_args_list[1][0][0]
    assert download_pkgs[0].checksum == "2111111111111111111111111111111111111111111111111111111111111111"
    assert download_pkgs[0].checksum_type == librepo.SHA256
    assert download_pkgs[0].relative_url == "Packages/b/b"
    assert download_pkgs[0].dest == "cachedir/sha256:2111111111111111111111111111111111111111111111111111111111111111"
    assert download_pkgs[0].handle.mirrorlisturl == "http://example.com/mirrorlist2"


class FakeSubscriptionManager:
    def __init__(self):
        self.get_secrets_calls = []

    def get_secrets(self, urls: List[str]):
        self.get_secrets_calls.append(urls)
        return {
            "ssl_ca_cert": "rhsm-ca-cert",
            "ssl_client_cert": "rhsm-client-cert",
            "ssl_client_key": "rhsm-client-key",
        }


@patch("librepo.download_packages")
def test_librepo_secrets_rhsm(mocked_download_pkgs, sources_service):
    TEST_SOURCES = {
        "sha256:1111111111111111111111111111111111111111111111111111111111111111": {
            "path": "Packages/a/a",
            "mirror": "mirror_id",
        }
    }
    sources_service.options = {
        "mirrors": {
            "mirror_id": {
                "url": "http://example.com/mirrorlist",
                "type": "mirrorlist",
                "secrets": {
                    "name": "org.osbuild.rhsm",
                }
            }
        }
    }
    sources_service.cache = "cachedir"
    sources_service.subscriptions = FakeSubscriptionManager()
    sources_service.fetch_all(TEST_SOURCES)
    assert len(mocked_download_pkgs.call_args_list) == 1
    # extract the list of packages to download (all_calls->call()->args->args[0])
    download_pkgs = mocked_download_pkgs.call_args_list[0][0][0]
    assert download_pkgs[0].checksum == "1111111111111111111111111111111111111111111111111111111111111111"
    assert download_pkgs[0].handle.sslclientkey == "rhsm-client-key"
    assert download_pkgs[0].handle.sslclientcert == "rhsm-client-cert"
    assert download_pkgs[0].handle.sslcacert == "rhsm-ca-cert"
    # double check that get_secrets() was called
    assert sources_service.subscriptions.get_secrets_calls == [["http://example.com/mirrorlist"]]


@patch("librepo.download_packages")
def test_librepo_secrets_mtls(mocked_download_pkgs, sources_service, monkeypatch):
    TEST_SOURCES = {
        "sha256:1111111111111111111111111111111111111111111111111111111111111111": {
            "path": "Packages/a/a",
            "mirror": "mirror_id",
        }
    }
    sources_service.options = {
        "mirrors": {
            "mirror_id": {
                "url": "http://example.com/mirrorlist",
                "type": "mirrorlist",
                "secrets": {
                    "name": "org.osbuild.mtls",
                }
            }
        }
    }
    monkeypatch.setenv("OSBUILD_SOURCES_CURL_SSL_CLIENT_KEY", "mtls-client-key")
    monkeypatch.setenv("OSBUILD_SOURCES_CURL_SSL_CLIENT_CERT", "mtls-client-cert")
    monkeypatch.setenv("OSBUILD_SOURCES_CURL_SSL_CA_CERT", "mtls-ca-cert")

    sources_service.cache = "cachedir"
    sources_service.fetch_all(TEST_SOURCES)
    assert len(mocked_download_pkgs.call_args_list) == 1
    # extract the list of packages to download (all_calls->call()->args->args[0])
    download_pkgs = mocked_download_pkgs.call_args_list[0][0][0]
    assert download_pkgs[0].checksum == "1111111111111111111111111111111111111111111111111111111111111111"
    assert download_pkgs[0].handle.sslclientkey == "mtls-client-key"
    assert download_pkgs[0].handle.sslclientcert == "mtls-client-cert"
    assert download_pkgs[0].handle.sslcacert == "mtls-ca-cert"


@pytest.mark.parametrize("test_mirrors,expected_err", [
    # bad
    (
        # no spaces in mirror_id
        {"bad mirror": {"url": "http://example.com", "type": "mirrorlist"}},
        "'bad mirror' does not match any of the regexes: ",
    ),
    (
        {"0123456789abcdef": {"type": "mirrorlist"}},
        "'url' is a required property",
    ),
    (
        {"0123456789abcdef": {"url": "http://example.com"}},
        "'type' is a required property",
    ),
    (
        {"0123456789abcdef": {"url": "http://example.com", "type": "bad_type"}},
        "'bad_type' is not one of ['mirrorlist', 'metalink', 'baseurl']",
    ),
    # good
    (
        {}, "",
    ),
    (
        {"0123": {"url": "http://example.com", "type": "mirrorlist"}}, "",
    ),
    (
        {"0123": {"url": "http://example.com", "type": "metalink"}}, "",
    ),
    (
        {"0123": {"url": "http://example.com", "type": "baseurl"}}, "",
    ),
    (
        {"Com_plex:id.": {"url": "http://x.com", "type": "baseurl"}}, "",
    ),
])
def test_schema_validation(sources_schema, test_mirrors, expected_err):
    test_input = {
        "items": {
            "sha256:2111111111111111111111111111111111111111111111111111111111111111": {
                "path": "Packages/b/b",
                "mirror": "mirror_id",
            },
        },
        "options": {
            "mirrors": test_mirrors,
        }
    }
    res = sources_schema.validate(test_input)
    if expected_err == "":
        assert res.valid
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err)


def test_librepo_copy_all(tmp_path, sources_service):
    src_cache = tmp_path / "src-cache"
    dst_cache = tmp_path / "dst-cache"
    fake_httpd_root = tmp_path / "fake-httpd-root"
    os.mkdir(src_cache)
    os.mkdir(dst_cache)
    os.mkdir(fake_httpd_root)

    TEST_SOURCES = {
        "sha256:ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb": {
            "path": "Packages/a/a",
            "mirror": "mirror_id",
        },
        "sha256:3e23e8160039594a33894f6564e1b1348bbd7a0088d42c4acb73eeaed59c009d": {
            "path": "Packages/b/b",
            "mirror": "mirror_id",
        },
    }

    for source in TEST_SOURCES.values():
        path = source["path"]
        os.makedirs(fake_httpd_root / os.path.dirname(path))

        basename = os.path.basename(source["path"])
        (fake_httpd_root / path).write_text(basename)

    with http_serve_directory(fake_httpd_root) as httpd:
        sources_service.setup(
            {
                "cache": src_cache,
                "options": {
                    "mirrors": {
                        "mirror_id": {
                            "url": f"http://localhost:{httpd.server_port}",
                            "type": "baseurl",
                        },
                    }
                },
            },
        )
        sources_service.fetch_all(TEST_SOURCES)
        assert httpd.reqs.count == len(TEST_SOURCES)

    for source in TEST_SOURCES:
        assert sources_service.exists(source, None)

    sources_service.copy_all(TEST_SOURCES, dst_cache)

    for source in TEST_SOURCES:
        assert os.path.isfile(dst_cache / sources_service.content_type / source)
